import numpy as np
import igraph as ig

def normalize_angles(angles):
    """
    reduce angles (in radians) to +/- pi.

    Parameters
    ----------
    angles : ArrayLike
        Angles to noramlize/reduce

    Returns
    -------
    angles : ArrayLike
        Reduced/normalized angles
    """

    return np.arctan2(np.sin(angles), np.cos(angles))


def angle_dist(angle1, angle2):
    """
    Get the minimum difference (distance) between two angles

    Parameters
    ----------
    angle1 : ArrayLike
        first set of angles (in radians)

    angle2: ArrayLike
        second set of angles (in radians)

    Returns
    -------
    distance : ArrayLike
        The angular distance (in radians) between angle1 and angle2 on the unit circle
    """
    diff = angle1 - angle2
    return np.arctan2(np.sin(diff), np.cos(diff))


def simple_cycle_vertices(graph: ig.Graph):
    """
    Find the vertices of all simple cycles in the graph

    Parameters
    ----------
    graph: ig.Graph
        The graph to find the cycle vertices of.

    Returns
    -------
    found_cycle_nodes : List[List[int]]
        A list of all simple cycles in the graph. Each simple cycle is composed of a list of vertices that belong to
        that cycle.
    """

    found_cycle_edges = graph.fundamental_cycles()
    found_cycle_nodes = []

    for cycle in found_cycle_edges:
        cyverts = set()
        for edge in graph.es(cycle):
            cyverts.update(edge.tuple)

        found_cycle_nodes.append(sorted(cyverts))
    return found_cycle_nodes


import numpy as np
from dataclasses import dataclass
from chilife.numba_utils import _ic_to_cart

class ProteinIC:

    def __init__(self, zmats, zmat_idxs, atom_dict, ICs, **kwargs):
        self.zmats = zmats
        self.zmat_idxs = zmat_idxs
        self.atom_chains = np.concatenate([[key] * len(zmats[key]) for key in zmats])
        self.atom_dict = atom_dict
        self.ICs = ICs
        self.atoms = np.asarray(kwargs['atoms']) if 'atoms' in kwargs else np.array([ic for chain in self.ICs.values()
                                                                                     for resi in chain.values()
                                                                                  for ic in resi.values()])

        self.atom_types = np.asarray(kwargs['atom_types']) if 'atom_types' in kwargs else \
            np.array([atom.atype for atom in self.atoms])
        self.atom_names = np.asarray(kwargs['atom_names']) if 'atom_names' in kwargs else \
            np.array([atom.name for atom in self.atoms])

        self.resis = np.array([key for i in self.ICs for key in self.ICs[i]])
        self.resnames = {
            j: next(iter(self.ICs[i][j].values())).resn
            for i in self.ICs
            for j in self.ICs[i]
        }
        self.chains = np.array([name for name in ICs])

        self.chain_operators = kwargs.get("chain_operators", None)
        self.has_chain_operators = True
        self.bonded_pairs = np.array(kwargs.get("bonded_pairs", None))
        self._nonbonded_pairs = kwargs.get('nonbonded_pairs', None)

        self.perturbed = False
        self._coords = kwargs['coords'] if 'coords' in kwargs else self.to_cartesian()
        self._dihedral_defs = None

    def copy(self):
        """Create a deep copy of an MolSysIC instance"""
        zmats = {key: value.copy() for key, value in self.zmats.items()}
        zmat_idxs = {key: value.copy() for key, value in self.zmat_idxs.items()}
        kwargs = {"chain_operators": self.chain_operators,
                  'atoms': self.atoms.copy(),
                  'atom_types': self.atom_types,
                  'atom_names': self.atom_names,
                  "bonded_pairs": self.bonded_pairs,
                  "_nonbonded_pairs": self._nonbonded_pairs,
                  'coords': self.coords.copy()}

        return ProteinIC(zmats, zmat_idxs, self.atom_dict, self.ICs, **kwargs)



    @property
    def coords(self):
        if (self._coords is None) or self.perturbed:
            self._coords = self.to_cartesian()
            self.perturbed = False
        return self._coords

    def get_dihedral(self, resi, atom_list, chain=None):

        if len(atom_list) == 0:
            return np.array([])

        chain = self._check_chain(chain)

        atom_list = np.atleast_2d(atom_list)
        dihedrals = []
        for atoms in atom_list:
            stem, stemr = tuple(atoms[2::-1]), tuple(atoms[1:])
            if (resi, stem) in self.atom_dict['dihedrals'][chain]:
                aidx = self.atom_dict['dihedrals'][chain][resi, stem][atoms[-1]]
            elif (resi, stemr) in self.atom_dict['dihedrals'][chain]:
                aidx = self.atom_dict['dihedrals'][chain][resi, stemr][atoms[0]]
            else:
                raise ValueError(
                    f"Dihedral with atoms {atoms} not found in chain {chain} on resi {resi} internal coordinates:\n"
                    + "\n".join([str(ic) for ic in self.ICs[chain][resi]])
                )
            dihedrals.append(self.zmats[chain][aidx, 2])

        return dihedrals[0] if len(dihedrals) == 1 else np.array(dihedrals)

    def get_dihedral_idx(self, resi, atom_list, chain=None):
        if len(atom_list) == 0:
            return np.array([])

        chain = self._check_chain(chain)

        atom_list = np.atleast_2d(atom_list)
        idxs = []
        for atoms in atom_list:
            stem, stemr = tuple(atoms[2::-1]), tuple(atoms[1:])
            if (resi, stem) in self.atom_dict['dihedrals'][chain]:
                aidx = self.atom_dict['dihedrals'][chain][resi, stem][atoms[-1]]
            elif (resi, stemr) in self.atom_dict['dihedrals'][chain]:
                aidx = self.atom_dict['dihedrals'][chain][resi, stemr][atoms[0]]
            else:
                raise ValueError(
                    f"Dihedral with atoms {atoms} not found in chain {chain} on resi {resi} internal coordinates:\n"
                    + "\n".join([str(ic) for ic in self.ICs[chain][resi]])
                )
            idxs.append(aidx)

        return idxs[0] if len(idxs) == 1 else np.array(idxs)

    def has_clashes(self, distance=1.5) -> bool:

        diff = (
            self.coords[self.nonbonded_pairs[:, 0]]
            - self.coords[self.nonbonded_pairs[:, 1]]
        )
        dist = np.linalg.norm(diff, axis=1)
        has_clashes = np.any(dist < distance)
        return has_clashes

    def get_resi_dihs(self, resi):

        if resi == 0 or resi == list(self.ICs[1].keys())[-1]:
            dihedral_defs = []
        elif resi == 1:
            dihedral_defs = [
                ["CH3", "C", "N", "CA"],
                ["C", "N", "CA", "C"],
                ["N", "CA", "C", "N"],
            ]
        else:
            dihedral_defs = [["C", "N", "CA", "C"], ["N", "CA", "C", "N"]]
        return dihedral_defs

    @property
    def dihedral_defs(self):
        if self._dihedral_defs is None:
            self._dihedral_defs = self.collect_dih_list()
        return self._dihedral_defs


    def shift_resnum(self, delta: int):
        for key in self.atom_dict:
            for chain in self.atom_dict[key]:

                self.atom_dict[key][chain] = {(res + delta, other): self.atom_dict[key][chain][res, other]
                                              for res, other in self.atom_dict[key][chain]}


        for segid in self.ICs:
            for resi in list(self.ICs[segid].keys()):
                if abs(delta) > 0:
                    self.ICs[segid][resi + delta] = self.ICs[segid][resi]
                    del self.ICs[segid][resi]

                for atom in self.ICs[segid][resi + delta].values():
                    atom.resi += delta
                    atom.dihedral_resi += delta

        self.resis = [key for i in self.ICs for key in self.ICs[i]]

    def phi_idxs(self, resnums = None, chain = None):
        chain = self._check_chain(chain)
        idxs =  np.argwhere((self.atom_names == 'C') * (self.atom_chains == chain)).flatten()
        if resnums is not None:
            resnums = np.atleast_1d(resnums)
            idxs = [self.atoms[idx].index for idx in idxs if (self.atoms[idx].resi in resnums)]
        idxs = np.asarray(idxs)
        if len(idxs) > 0:
            idxs = idxs[~np.isnan(self.zmats[chain][idxs, 2])]
        return idxs

    def psi_idxs(self, resnums = None, chain = None):
        chain = self._check_chain(chain)
        idxs = np.argwhere((self.atom_names == 'N')).flatten()
        if resnums is not None:
            resnums = np.atleast_1d(resnums) + 1
            idxs = [self.atoms[idx].index for idx in idxs if (self.atoms[idx].resi in resnums)]
        idxs = np.asarray(idxs)

        if len(idxs) > 0:
            idxs = idxs[~np.isnan(self.zmats[chain][idxs, 2])]
        return idxs

    def bb_idxs(self,  resnums= None, chain = None):
        return np.concatenate((self.phi_idxs(resnums, chain), self.psi_idxs(resnums, chain)))

    def _check_chain(self, chain):
        if chain is None and len(self.ICs) == 1:
            chain = list(self.ICs.keys())[0]

        elif chain is None and len(self.ICs) > 1:
            raise ValueError("You must specify the protein chain")

        return chain


    #===============================================================================================================
    #                              KNOWN NECESSARY
    #===============================================================================================================

    def to_cartesian(self):
        """Convert internal coordinates into cartesian coordinates.

        Returns
        -------
        coords_array : np.ndarray
            Array of cartesian coordinates corresponding to ICAtom list atoms
        """
        coord_arrays = []
        for segid in self.chain_operators:
            # Prepare variables for numba compiled function

            IC_idx_array, ICArray = self.zmat_idxs[segid], self.zmats[segid]
            cart_coords = _ic_to_cart(IC_idx_array, ICArray)

            # Apply chain operations if any exist
            if hasattr(self, 'has_chain_operators'):
                if self.has_chain_operators:
                    cart_coords = cart_coords @ self.chain_operators[segid]["mx"] + self.chain_operators[segid]["ori"]

            coord_arrays.append(cart_coords)

        return np.concatenate(coord_arrays)

    @property
    def chain_operators(self):
        return self._chain_operators

    @chain_operators.setter
    def chain_operators(self, op):
        if op is None:
            op = {
                chain: {"ori": np.array([0, 0, 0]), "mx": np.identity(3)}
                for chain in self.ICs
            }
            self.has_chain_operators = False
        else:
            self.has_chain_operators = True
        self._chain_operators = op
        self.perturbed = True

@dataclass
class ICAtom:
    name: str
    atype: str
    index: int
    resn: str
    resi: int
    atom_names: tuple

    bond_idx: int = np.nan
    bond: float = np.nan
    angle_idx: int = np.nan
    angle: float = np.nan
    dihedral_idx: int = np.nan
    dihedral: float = np.nan
    dihedral_resi: int = np.nan

    def __post_init__(self):
        """If no dihedral_resi is defined default to the atom residue"""
        if np.isnan(self.dihedral_resi):
            self.dihedral_resi: int = self.resi


import os, hashlib
import numpy as np
import chilife as xl
import MDAnalysis as mda

ubq = mda.Universe('test_data/1ubq.pdb')
r = np.linspace(15, 70, 256)


def test_creation():
    sele = ubq.select_atoms('resnum 59')

    # Spin density -- Bleifuss et al. Biochemistry 2001. -- Why don't the spin densities sum to 1? IDK Ask Bleifuss.
    ISL = xl.IntrinsicLabel("TYX", sele, spin_atoms={'CG': 0.38, 'CE1': 0.25, 'CE2': 0.25, 'OH': 0.29})
    SL1 = xl.SpinLabel('R1M', 28, ubq)
    P = xl.distance_distribution(ISL, SL1, r)

    assert r[np.argmax(P)] - 21.470588235294116 < 1e-6
    assert np.max(P) - 0.25610254327499965 < 1e-6


def test_save():
    sele = ubq.select_atoms('resnum 59')
    ISL = xl.IntrinsicLabel("TYX", sele, spin_atoms={'CG': 0.38, 'CE1': 0.25, 'CE2': 0.25, 'OH': 0.29})
    xl.save('test_data/ISL_test.pdb', ISL)

    with open('test_data/ISL.pdb', 'r') as f:
        ans = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    with open('test_data/ISL_test.pdb', 'r') as f:
        test = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    os.remove('test_data/ISL_test.pdb')

    assert test == ans


def test_single_atom():
    azurin = xl.fetch('1e67')
    Zn = azurin.select_atoms('segid D and resnum 129')
    ISL = xl.IntrinsicLabel("ZCU", Zn, spin_atoms="ZN")
    np.testing.assert_allclose(ISL.spin_centers, [[ 5.59399986, 31.97200012, 39.02500153]])


def test_ag_spin_atoms():
    azurin = xl.fetch('1e67')
    Zn = azurin.select_atoms('segid D and resnum 129')
    ISL = xl.IntrinsicLabel("ZCU", Zn, spin_atoms=Zn)
    np.testing.assert_allclose(ISL.spin_centers, [[ 5.59399986, 31.97200012, 39.02500153]])


from __future__ import annotations
import numbers
import operator
from functools import partial, update_wrapper

import MDAnalysis
from MDAnalysis.core.selection import AtomSelection

from .pdb_utils import sort_pdb
from .Topology import Topology
import numpy as np
from numpy.typing import ArrayLike
from scipy.spatial import cKDTree
import chilife


# TODO:
#   Behavior: AtomSelections should have orders to be enforced when indexing.

masked_properties = ('record_types', 'atomids', 'names', 'altlocs', 'altLocs', 'resnames', 'resnums', 'icodes', 'resids',
                     'resis', 'chains', 'occupancies', 'bs', 'segs', 'segids', 'atypes', 'types', 'charges', 'ix',
                     'resixs', 'segixs', '_Atoms', 'atoms')

singles = ('record_type', 'name', 'altloc', 'altLoc', 'atype', 'type', 'resn', 'resname', 'resnum', 'resid', 'icode',
           'resi', 'chain', 'segid', 'charge')


class MolecularSystemBase:
    """Base class for molecular systems containing attributes universal to """

    def __getattr__(self, key):
        if 'molsys' not in self.__dict__:
            self.__getattribute__(key)
        elif key in singles:
            return np.squeeze(self.molsys.__getattribute__(key + 's')[self.mask]).flat[0]
        elif key not in self.__dict__['molsys'].__dict__:
            self.molsys.__getattribute__(key)
        elif key == 'trajectory':
            return self.molsys.trajectory
        elif key == 'atoms':
            return self.molsys.__getattribute__(key)[self.mask]
        elif key in masked_properties or key in singles:
            return np.squeeze(self.molsys.__getattribute__(key)[self.mask])
        else:
            return self.molsys.__getattribute__(key)

    def __setattr__(self, key, value):
        if key in ('molsys', 'mask'):
            super(MolecularSystemBase, self).__setattr__(key, value)
        elif key in singles:
            self.molsys.__getattribute__(key + 's')[self.mask] = value
        elif key not in self.__dict__['molsys'].__dict__:
            super(MolecularSystemBase, self).__setattr__(key, value)
        elif key == 'trajectory':
            self.molsys.__dict__['trajectory'] = value
        elif key in masked_properties or key in singles:
            self.molsys.__getattribute__(key)[self.mask] = value
        else:
            super(MolecularSystemBase, self).__setattr__(key, value)

    def select_atoms(self, selstr):
        """
        Select atoms from :class:`~MolSys` or :class:`~AtomSelection` based on selection a selection string, similar
        to the `select_atoms` method from `MDAnalysis <https://docs.mdanalysis.org/stable/documentation_pages/selections.html>`_

        Parameters
        ----------
        selstr : str
            Selection string

        Returns
        -------
        atoms : :class:`~AtomSelection`
            Selected atoms
        """

        mask = process_statement(selstr, self.logic_keywords, self.molsys_keywords)
        if hasattr(self, 'mask'):
            t_mask = np.zeros(self.molsys.n_atoms, dtype=bool)
            t_mask[self.mask] = True
            mask *= t_mask

        mask = np.argwhere(mask).T[0]
        return AtomSelection(self.molsys, mask)

    def __getitem__(self, item):

        if np.issubdtype(type(item), np.integer):
            relidx = self.molsys.ix[self.mask][item]
            return Atom(self.molsys, relidx)

        if isinstance(item, slice):
            return AtomSelection(self.molsys, self.mask[item])

        elif isinstance(item, (np.ndarray, list, tuple)):
            item = np.asarray(item)
            if len(item) == 1 or item.sum() == 1:
                return Atom(self.molsys, self.mask[item])
            elif item.dtype == bool:
                item = np.argwhere(item).T[0]
                return AtomSelection(self.molsys, item)
            else:
                return AtomSelection(self.molsys, self.mask[item])

        elif hasattr(item, '__iter__'):
            if all([np.issubdtype(type(x), int) for x in item]):
                return AtomSelection(self.molsys, self.mask[item])

        raise TypeError('Only integer, slice type, and boolean mask arguments are supported at this time')


    @property
    def fname(self):
        """File name or functional name of the molecular system"""
        return self.molsys._fname

    @fname.setter
    def fname(self, val):
        self.molsys._fname = val

    @property
    def positions(self):
        """3-Dimensional cartesian coordinates of the atoms in the molecular system"""
        return self.coords

    @positions.setter
    def positions(self, val):
        self.coords = np.asarray(val)

    @property
    def coords(self):
        """3-dimensional cartesian coordinates of the atoms in the molecular system"""
        return self.trajectory.coords[self.trajectory.frame, self.mask]

    @coords.setter
    def coords(self, val):
        self.trajectory.coords[self.trajectory.frame, self.mask] = np.asarray(val)

    @property
    def n_atoms(self):
        """"Number of atoms in the molecular system"""
        return len(self.mask)

    @property
    def n_residues(self):
        """Number of residues in the molecular system"""
        return len(np.unique(self.resixs[self.mask]))

    @property
    def n_chains(self):
        """Number of chains in the molecular system"""
        return len(np.unique(self.chains[self.mask]))

    @property
    def residues(self):
        """:class:`~ResidueSelection` for all residues of the molecular system"""
        return ResidueSelection(self.molsys, self.mask)

    @property
    def segments(self):
        """:class:`~SegmentSelection` for all segments (chains) of the molecular system"""
        return SegmentSelection(self.molsys, self.mask)

    @property
    def logic_keywords(self):
        """Dictionary of logic keywords (strings) pertaining to :meth:`atom_selection` method. Keywords are mapped
        to the associated logic function """
        return self.molsys._logic_keywords

    @property
    def molsys_keywords(self):
        """Dictionary of molecular system keywords (strings) pertaining to :meth:`atom_selection` method. Keywords are
        mapped to the associated molecular system attributes"""
        return self.molsys._molsys_keywords


    @property
    def universe(self):
        """Wrapper attribute to allow molecular systems to behave like MDAnalysis AtomGroup objects"""
        return self.molsys

    def copy(self):
        """Creates a deep copy of the underlying MolSys and returns an :class:`AtomSelection` from the new copy
        matching the current selection if it exists."""
        p2 = self.molsys.copy()
        return AtomSelection(p2, self.mask.copy())

    def __iter__(self):
        """Iterate over all atoms of the MolSys"""
        for idx in self.mask:
            yield self.molsys.atoms[idx]

    def __eq__(self, value):
        """Checks if two molecular systems are equal by making sure they have the same ``self.molsys`` and the same
        mask"""
        return self.molsys is value.molsys and self.mask == value.mask



class MolSys(MolecularSystemBase):
    """
    An object containing all attributes of a molecular system.

    Parameters
    ----------
    atomids : np.ndarray
        Array of atom indices where each index is unique to each atom in the molecular system.
    names : np.ndarray
        Array of atom names, e.g. 'N', 'CA' 'C'.
    altlocs : np.ndarray
        Array of alternative location identifiers for each atom.
    resnames : np.ndarray
        Array of residue names for each atom.
    resnums : np.ndarray
        Array of residue numbers for each atom.
    chains : np.ndarray
        Array of chain identifiers for each atom.
    trajectory : np.ndarray
        Cartesian coordinates of each atom for each state of an ensemble or for each timestep of a trajectory.
    occupancies : np.ndarray
        Occupancy or q-factory for each atom.
    bs : np.ndarray
        thermal, or b-factors for each atom.
    segs : np.ndarray
        Segment/chain IDs for each atom.
    atypes : np.ndarray
        Atom/element types.
    charges : np.ndarray
        Formal or partial charges assigned to each atom.
    bonds : ArrayLike
        Array of atom ID pairs corresponding to all atom bonds in the system.
    name : str
        Name of molecular system
    """
    def __init__(
            self,
            record_types: np.ndarray,
            atomids: np.ndarray,
            names: np.ndarray,
            altlocs: np.ndarray,
            resnames: np.ndarray,
            resnums: np.ndarray,
            icodes: np.ndarray,
            chains: np.ndarray,
            trajectory: np.ndarray,
            occupancies: np.ndarray,
            bs: np.ndarray,
            segs: np.ndarray,
            atypes: np.ndarray,
            charges: np.ndarray,
            bonds: ArrayLike = None,
            name: str = 'Noname_MolSys'

    ):

        self.molsys = self
        self.record_types = record_types.copy()
        self.atomids = atomids.copy()
        self.names = names.copy().astype('U4')
        self.altlocs = altlocs.copy()
        self.resnames = resnames.copy()
        self.resnums = resnums.copy()
        self.icodes = icodes.copy()
        self.chains = chains.copy()
        self.trajectory = Trajectory(trajectory.copy(), self)
        self.occupancies = occupancies.copy()
        self.bs = bs.copy()
        self.segs = segs.copy()
        self.atypes = atypes.copy()
        self.charges = charges.copy()
        self._fname = name

        self.ix = np.arange(len(self.atomids))
        self.mask = np.arange(len(self.atomids))

        # resix_borders = np.nonzero(np.r_[1, np.diff(self.resnums)[:-1]])
        # resix_borders = np.append(resix_borders, [self.n_atoms])
        # resixs = []
        # for i in range(len(resix_borders) - 1):
        #     dif = resix_borders[i + 1] - resix_borders[i]
        #     resixs.append(np.ones(dif, dtype=int) * i)

        residues, resixs = [], []
        i = 0
        for aidx in range(self.n_atoms):
            residue_id = self.resnums[aidx], self.icodes[aidx], self.chains[aidx]
            if residue_id not in residues:
                residues.append(residue_id)
                i += 1

            resixs.append(i)

        self.resixs = np.array(resixs)
        self.resindices = self.resixs
        if np.all(self.chains == '') or np.all(self.chains == 'SYSTEM'):
            self.segixs = np.array([ord('A') - 65 for x in self.chains])
        else:
            self.segixs = np.array([ord(x) - 65 for x in self.chains])
        self.segindices = self.segixs

        self.is_protein = np.array([res in chilife.SUPPORTED_RESIDUES for res in resnames])

        self._molsys_keywords = {'id': self.atomids,
                                 'record_type': self.record_types,
                                 'name': self.names,
                                 'altloc': self.altlocs,
                                 'resname': self.resnames,
                                 'resnum': self.resnums,
                                 'icode': self.icodes,
                                 'chain': self.chains,
                                 'occupancies': self.occupancies,
                                 'b': self.bs,
                                 'segid': self.chains,
                                 'type': self.atypes,
                                 'charges': self.charges,

                                 'resi': self.resnums,
                                 'resid': self.resnums,
                                 'i.': self.resnums,
                                 's.': self.chains,
                                 'c.': self.chains,
                                 'r.': self.resnames,
                                 'n.': self.names,
                                 'resn': self.resnames,
                                 'q': self.occupancies,
                                 'elem': self.atypes,
                                 'element': self.atypes,

                                 'protein': self.is_protein,
                                 '_len': self.n_atoms}

        self._logic_keywords = {'and': operator.mul,
                                'or': operator.add,
                                'not': unot,
                                '!': unot,
                                '<': operator.lt,
                                '>': operator.gt,
                                '==': operator.eq,
                                '<=': operator.le,
                                '>=': operator.ge,
                                '!=': operator.ne,
                                'byres': partial(byres, molsys=self.molsys),
                                'within': update_wrapper(partial(within, molsys=self.molsys), within),
                                'around': update_wrapper(partial(within, molsys=self.molsys), within)}


        # Aliases
        self.resns = self.resnames
        self.resis = self.resnums
        self.altLocs = self.altlocs
        self.resids = self.resnums
        self.segids = self.chains
        self.types = self.atypes

        # Atoms creation is required to be last
        self.atoms = AtomSelection(self, self.mask)
        self.topology = Topology(self, bonds) if bonds is not None else None


    @classmethod
    def from_pdb(cls, file_name, sort_atoms=False):
        """
        Reads a pdb file and returns a MolSys object

        Parameters
        ----------
        file_name : str, Path
            Path to the PDB file to read.
        sort_atoms : bool
            Presort atoms for consisted internal coordinate definitions

        Returns
        -------
        cls : MolSys
            Molecular system object of the PDB structure.
        """

        keys = ["record_types", "atomids", "names", "altlocs", "resnames", "chains", "resnums",
                "icodes", "coords", "occupancies", "bs", "segs", "atypes", "charges"]
        if sort_atoms:
            lines = sort_pdb(file_name)
        else:
            with open(file_name, 'r') as f:
                lines = f.readlines()

            lines = [line for line in lines if line.startswith(('MODEL', 'ENDMDL', 'ATOM', 'HETATM'))]
            start_idxs, end_idxs = [],  []
            for i, line in enumerate(lines):
                if line.startswith('MODEL'):
                    start_idxs.append(i + 1)
                elif line.startswith("ENDMDL"):
                    end_idxs.append(i)

            if len(start_idxs) > 0:
                lines = [lines[start:end] for start, end in zip(start_idxs, end_idxs)]

        if isinstance(lines[0], str):
            lines = [lines]

        PDB_data = [(line[:6].strip(), i, line[12:16].strip(), line[16:17].strip(),
                     line[17:20].strip(), line[21:22].strip(), int(line[22:26]), line[26:27].strip(),
                     (float(line[30:38]), float(line[38:46]), float(line[46:54])), float(line[54:60]),
                     float(line[60:66]), line[72:73].strip(), line[76:78].strip(), line[78:80].strip())
                    for i, line in enumerate(lines[0])]

        pdb_dict = {key: np.array(data) for key, data in zip(keys, zip(*PDB_data))}
        trajectory = [pdb_dict.pop('coords')]

        if len(lines) > 1:
            for struct in lines[1:]:
                frame_coords = [(line[30:38], line[38:46], line[46:54]) for line in struct]

                if len(frame_coords) != len(PDB_data):
                    raise ValueError('All models in a multistate PDB must have the same atoms')

                trajectory.append(frame_coords)

        pdb_dict['trajectory'] = np.array(trajectory, dtype=float)
        pdb_dict['name'] = file_name
        return cls(**pdb_dict)

    @classmethod
    def from_arrays(cls,
                    anames: ArrayLike,
                    atypes: ArrayLike,
                    resnames: ArrayLike,
                    resindices: ArrayLike,
                    resnums: ArrayLike,
                    segindices: ArrayLike,
                    record_types: ArrayLike = None,
                    altlocs: ArrayLike = None,
                    icodes: ArrayLike = None,
                    segids: ArrayLike = None,
                    trajectory: ArrayLike = None,
                    **kwargs
                    ) -> MolSys:
        """
        Create a MolSys object from a minimal set of arrays.

        Parameters
        ----------
        anames : ArrayLike
            Array of atom names.
        atypes : ArrayLike
            Array of atom types/elements.
        resnames : ArrayLike
            Array of residue names for each atom.
        resindices : ArrayLike
            Array of residue indices for each atom.
        resnums : ArrayLike
            Array of residue numbers for each atom.
        segindices : ArrayLike
            Array of segment indices for each atom.
        segids : ArrayLike
            Array of segment IDs for each atom.
        trajectory : ArrayLike
            Array of cartesian coordinates for each atom and each state of an ensemble or frame of a trajectory.

        Returns
        -------
        cls : MolSys
            Molecular system object of the PDB structure.
        """
        anames = np.asarray(anames)
        atypes = np.asarray(atypes)
        resindices = np.asarray(resindices)
        segindices = np.asarray(segindices)

        n_atoms = len(anames)
        atomids = np.arange(n_atoms)
        altlocs = np.array([''] * n_atoms) if altlocs is None else np.array(altlocs)
        icodes = np.array([''] * n_atoms) if icodes is None else np.array(icodes)


        if len(resnums) != n_atoms:
            resnums = np.array([resnums[residx] for residx in resindices])
        else:
            resnums = np.asarray(resnums)

        if len(resnames) != n_atoms:
            resnames = np.array([resnames[residx] for residx in resindices])
        else:
            resnames = np.asarray(resnames)

        if len(icodes) != n_atoms:
            icodes = np.array([icodes[residx] for residx in resindices])
        else:
            icodes = np.asarray(icodes)

        if len(segindices) != n_atoms:
            segindices = np.array([segindices[x] for x in resindices])

        if segids is None:
            segids = np.array(["ABCDEFGHIJKLMNOPQRSTUVWXYZ"[i] for i in segindices])
        elif len(segids) != n_atoms:
            segids = np.array([segids[i] for i in segindices])
        else:
            segids = np.asarray(segids)

        chains = segids

        if trajectory is None:
            trajectory = np.empty((1, n_atoms, 3))
        elif len(trajectory.shape) == 2:
            trajectory = trajectory[None, ...]

        if record_types is None:
            record_types = np.array(['ATOM' if rnme in chilife.SUPPORTED_RESIDUES else 'HETATM' for rnme in resnames])

        occupancies = np.ones(n_atoms)
        bs = np.ones(n_atoms)
        charges = kwargs.pop('charges', np.zeros(n_atoms))

        return cls(record_types, atomids, anames, altlocs, resnames, resnums, icodes, chains,
                   trajectory, occupancies, bs, segids, atypes, charges, **kwargs)

    @classmethod
    def from_atomsel(cls, atomsel, frames=None):
        """
        Create a new MolSys object from an existing :class:`~AtomSelection` or
        `MDAnalysis.AtomGroup <https://userguide.mdanalysis.org/stable/atomgroup.html>`_ object.

        Parameters
        ----------
        atomsel : :class:`~AtomSelection`, MDAnalysis.AtomGroup
            The :class:`~AtomSelection` or MDAnalysis.AtomGroup from which to create the new MolSys object.
        frames : int, Slice, ArrayLike
            Frame index, slice or array of frame indices corresponding to the frames of the atom selection you wish to
            extract into a new :class:`~MolSys` object.

        Returns
        -------
        cls : MolSys
            Molecular system object of the PDB structure.
        """

        if isinstance(atomsel, (MDAnalysis.AtomGroup, AtomSelection)):
            U = atomsel.universe
        elif isinstance(atomsel, MDAnalysis.Universe):
            U = atomsel
            atomsel = U.atoms

        if frames is None:
            frames = slice(0, len(U.trajectory))

        record_types = atomsel.record_types if hasattr(atomsel, 'record_types') else None
        icodes = atomsel.icodes if hasattr(atomsel, 'icodes') else None
        anames = atomsel.names
        atypes = atomsel.types
        resnames = atomsel.resnames
        resnums = atomsel.resnums
        ridx_map = {num: i for i, num in enumerate(np.unique(atomsel.resindices))}
        resindices = np.array([ridx_map[num] for num in atomsel.resindices ])
        segids = atomsel.segids
        sidx_map = {num: i for i, num in enumerate(np.unique(atomsel.segindices))}
        segindices = np.array([sidx_map[num] for num in atomsel.segindices])

        if hasattr(U.trajectory, 'coordinate_array'):
            trajectory = U.trajectory.coordinate_array[frames][:, atomsel.ix, :]
        else:
            trajectory = []
            for ts in U.trajectory[frames]:
                trajectory.append(atomsel.positions)
            trajectory = np.array(trajectory)

        return cls.from_arrays(anames, atypes, resnames, resindices, resnums, segindices,
                               segids=segids, trajectory=trajectory, record_types=record_types,
                               icodes=icodes)


    @classmethod
    def from_rdkit(cls, mol):
        """
        Create a MolSys from an rdkit Mol object with embedded conformers.

        Parameters
        ----------
        mol

        Returns
        -------

        """
        atypes = np.array([a.GetSymbol() for a in mol.GetAtoms()])
        anames = np.array([a + str(i) for i, a in enumerate(atypes)])
        resnames = np.array(["UNK" for _ in anames])
        resindices = np.array([0] * len(anames))
        resnums = np.array([1] * len(anames))
        segindices = np.array([0] * len(anames))
        segids = np.array(["A"] * len(anames))
        trajectory = np.array([conf.GetPositions() for conf in mol.GetConformers()])
        bonds = np.array([[b.GetBeginAtomIdx(), b.GetEndAtomIdx()] for b in mol.GetBonds()])
        return cls.from_arrays(anames, atypes, resnames, resindices, resnums, segindices,
                               segids=segids, trajectory=trajectory, bonds=bonds)

    @classmethod
    def from_openMM(cls, simulation):
        """
        Create a MolSys from an openMM Simulation object.

        Parameters
        ----------
        simulation : openmm.simulation.Simulation
            The simulation object to create the MolSys object from.

        Returns
        -------
        sys : chilife.MolSys
            chiLife molecular system as a clone from
        """
        try:
            from openmm.unit import angstrom
        except:
            raise RuntimeError("You must have the optional openMM dependency installed to use from_openMM")

        top = simulation.topology
        atypes = np.array([a.element.symbol for a in top.atoms()])
        anames = np.array([a.name for a in top.atoms()])
        resnames = np.array([a.residue.name for a in top.atoms()])
        resindices = np.array([a.residue.index for a in top.atoms()])
        resnums = np.array([int(a.residue.id) for a in top.atoms()])
        segindices = np.array([a.residue.chain.index for a in top.atoms()])
        segids = np.array([a.residue.chain.id for a in top.atoms()])
        trajectory = simulation.context.getState(getPositions=True).getPositions(asNumpy=True).value_in_unit(angstrom)
        bonds = np.array([[bond.atom1.index, bond.atom2.index] for bond in top.bonds()])
        sys = cls.from_arrays(anames, atypes, resnames, resindices, resnums,
                              segindices, segids=segids, trajectory=trajectory, bonds=bonds)
        return sys

    def copy(self):
        """Create a deep copy of the MolSys."""

        return MolSys(
            record_types=self.record_types,
            atomids=self.atomids,
            names=self.names,
            altlocs=self.altlocs,
            resnames=self.resnames,
            resnums=self.resnums,
            icodes=self.icodes,
            chains=self.chains,
            trajectory=self.trajectory.coords,
            occupancies=self.occupancies,
            bs=self.bs,
            segs=self.segs,
            atypes=self.atypes,
            charges=self.charges,
            name=self.names)

    def load_new(self, coordinates):
        """
        Load a new set of 3-dimensional coordinates into the MolSys.

        Parameters
        ----------
        coordinates : ArrayLike
            Set of new coordinates to load into the MolSys
        """
        self.trajectory.load_new(coordinates=coordinates)

    @property
    def _Atoms(self):
        if not hasattr(self, '__Atoms'):
            self.__Atoms = np.array([Atom(self, i) for i in range(self.n_atoms)])
        return self.__Atoms


class Trajectory:
    """
    Object containing and managing 3-dimensional coordinates of :class:`~MolSys` objects, particularly when there are
    multiple states or time steps.

    Parameters
    ----------
    coordinates : np.ndarray
        3-dimensional coordinates for all atoms and all states/frames of the associated :class:`~MolSys` object.
    molsys : MolSys
        The associated :class:`~MolSys` object.
    timestep : float
        The time step or change in time between frames/states of the trajectory/ensemble.
    """

    def __init__(self, coordinates : np.ndarray, molsys : MolSys, timestep : float=1):

        self.molsys = molsys
        self.timestep = timestep
        self.coords = coordinates
        self.coordinate_array = coordinates
        self._frame = 0
        self.time = np.arange(0, len(self.coords)) * self.timestep

    def load_new(self, coordinates):
        """
        Load a new set (trajectory or ensemble) of 3-dimensional coordinates into the Trajectory.

        Parameters
        ----------
        coordinates : ArrayLike
            Set of new coordinates to load into the MolSys
        """
        self.coords = coordinates
        self.coordinate_array = coordinates
        self.time = np.arange(0, len(self.coords)) * self.timestep

    def __getitem__(self, item):
        if isinstance(item, (slice, list, np.ndarray)):
            return TrajectoryIterator(self, item)

        elif isinstance(item, numbers.Integral):
            self._frame = item
            return self.time[self._frame]

    def __iter__(self):
        return iter(TrajectoryIterator(self))

    def __len__(self):
        return len(self.time)

    @property
    def frame(self):
        """The frame number the trajectory is currently on"""
        return self._frame

    @frame.setter
    def frame(self, value):
        self._frame = value


class TrajectoryIterator:
    """An iterator object for looping over :class:`~Trajectory` objects."""

    def __init__(self, trajectory, arg=None):
        self.trajectory = trajectory
        if arg is None:
            self.indices = range(len(self.trajectory.time))
        elif isinstance(arg, slice):
            self.indices = range(*arg.indices(len(self.trajectory.time)))
        elif isinstance(arg, (np.ndarray, list, tuple)):
            if np.issubdtype(arg.dtype, np.integer):
                self.indices = arg
            elif np.dtype == bool:
                self.indices = np.argwhere(arg)
            else:
                raise TypeError(f'{arg.dtype} is not a valid indexor for a trajectory')
        else:
            raise TypeError(f'{arg} is not a valid indexor for a trajectory')

    def __iter__(self):
        for frame in self.indices:
            yield self.trajectory[frame]

    def __len__(self):
        return len(self.indices)


def process_statement(statement, logickws, subjectkws):
    """
    Parse a selection string to identify subsets of atoms that satisfy the given conditions.

    Parameters
    ----------
    statement : str
        The selection string to parse.
    logickws : dict
        Dictionary of logic keywords (strings) mapped to the associated logic function.
    subjectkws : dict
        Dictionary of subject keywords (strings) mapped to the associated subject attributes.

    Returns
    -------
    mask : np.ndarray
        Array of subject indices that satisfy the given conditions.
    """
    sub_statements = parse_paren(statement)

    mask = np.ones(subjectkws['_len'], dtype=bool)
    operation = None

    for stat in sub_statements:
        if stat.startswith('('):
            tmp = process_statement(stat, logickws, subjectkws)
            mask = operation(mask, tmp) if operation else tmp
            continue

        stat_split = stat.split()
        subject = None
        values = []
        while len(stat_split) > 0:

            if stat_split[0] in subjectkws:
                subject = subjectkws[stat_split.pop(0)]
                values = []

            elif stat_split[0] in logickws:
                if subject is not None:
                    values = np.array(values, dtype=subject.dtype)
                    tmp = process_sub_statement(subject, values)
                    subject, values = None, []
                    mask = operation(mask, tmp) if operation else tmp

                # Get next operation
                operation, stat_split = build_operator(stat_split, logickws)

            else:
                values += parse_value(stat_split.pop(0))

        if subject is not None:
            values = np.array(values, dtype=subject.dtype)
            tmp = process_sub_statement(subject, values)
            mask = operation(mask, tmp) if operation else tmp
            operation = None

    return mask


def parse_paren(string):
    """
    Break down a given string to parenthetically defined subcomponents

    Parameters
    ----------
    string : str
        The string to break down.

    Returns
    -------
    results : List[str]
        List of parenthetically defined subcomponents.
    """
    stack = 0
    start_idx = 0
    end_idx = 0
    results = []

    for i, c in enumerate(string):
        if c == '(':
            if stack == 0:
                start_idx = i + 1
                if i != 0:
                    results.append(string[end_idx:i].strip())
            stack += 1
        elif c == ')':
            # pop stack
            stack -= 1

            if stack == 0:
                results.append(f'({string[start_idx:i].strip()})')
                start_idx = end_idx
                end_idx = i + 1

    if end_idx != len(string) - 1:
        results.append(string[end_idx:].strip())

    results = [res for res in results if res != '']
    if len(results) == 1 and results[0].startswith('('):
        results = parse_paren(results[0][1:-1])

    if stack != 0 :
        raise RuntimeError('The provided statement is missing a parenthesis or has an '
                           'extra one.')
    return results


def check_operation(operation, stat_split, logickws):
    """
    Given an advanced operation, identify the additional arguments being passed to the advanced operation and create
    a simple operation that defined specifically for the provided arguments.
    Parameters
    ----------
    operation : callable
        A function or callable object that operates on molecular system attributes.
    stat_split : List[str]
        the arguments associated with the operation
    logickws : dict
        Dictionary of logic keywords (strings) mapped to the associated logic function.

    Returns
    -------
    operation : callable
        A simplified version of the provided operation now accounting for the user provided parameters.
    """
    advanced_operators = (logickws['within'], logickws['around'], logickws['byres'])
    if operation in advanced_operators:
        outer_operation = logickws['and']
        args = [stat_split.pop(i) for i in range(1, 1 + operation.nargs)]
        operation = partial(operation, *args)


        def toperation(a, b, outer_operation, _io):
            return outer_operation(a, _io(b))

        operation = partial(toperation, outer_operation=outer_operation, _io=operation)

    return operation


def build_operator(stat_split, logickws):
    """
    A function to combine multiple sequential operators into one.

    Parameters
    ----------
    stat_split : List[str]
        The list of string arguments passed by the user to define the sequence of operations
    logickws : dict
        Dictionary of logic keywords (strings) mapped to the associated logic function.

    Returns
    -------
    operation : callable
        A compression of sequential operations combined into one.
    split_stat : List[str]
        Any remaining arguments passed by the user that were not used to define ``operation``
    """
    operation = logickws['and']
    unary_operators = (logickws['not'], logickws['byres'])
    binary_operators = (logickws['and'], logickws['or'])
    advanced_operators = (logickws['within'], logickws['around'])

    while _io := logickws.get(stat_split[0], False):

        if _io in unary_operators:
            def toperation(a, b, operation, _io):
                return operation(a, _io(b))

            operation = partial(toperation, operation=operation, _io=_io)
            stat_split = stat_split[1:]

        elif _io in advanced_operators:
            stat_split = stat_split[1:]
            args = [stat_split.pop(0) for i in range(_io.nargs)]
            _io = partial(_io, *args)

            def toperation(a, b, operation, _io):
                return operation(a, _io(b))

            operation = partial(toperation, operation=operation, _io=_io)

        elif _io in binary_operators:
            if operation != logickws['and']:
                raise RuntimeError('Cannot have two binary logical operators in succession')
            operation = _io
            stat_split = stat_split[1:]

        if len(stat_split) == 0:
            break

    return operation, stat_split


def parse_value(value):
    """
    Helper function to parse values that may have slices.

    Parameters
    ----------
    value : str, int, ArrayLike[int]
        Values to parse.

    Returns
    -------
    return_value : List
        A list of values that satisfy the input.

    """
    return_value = []
    if '-' in value:
        start, stop = [int(x) for x in value.split('-')]
        return_value += list(range(start, stop + 1))
    elif ':' in value:
        start, stop = [int(x) for x in value.split(':')]
        return_value += list(range(start, stop + 1))
    else:
        return_value.append(value)

    return return_value


def process_sub_statement(subject, values):

    if subject.dtype == bool:
        return subject

    return np.isin(subject, values)


class AtomSelection(MolecularSystemBase):
    """
    An object containing a group of atoms from a :class:`~MolSys` object

    Parameters
    ----------
    molsys : :class:`~MolSys`
        The :class:`~MolSys` object from which the AtomSelection belongs to.
    mask : np.ndarray
        An array of atom indices that defines the atoms of ``molsys`` that make up the atom selection.
    """
    def __new__(cls, *args):
        if len(args) == 2:
            molsys, mask = args
        else:
            return object.__new__(cls)

        if isinstance(mask, int):
            return Atom(molsys, mask)
        else:
            return object.__new__(cls)

    def __init__(self, molsys, mask):

        self.molsys = molsys
        self.mask = mask\

    def __getitem__(self, item):

        if np.issubdtype(type(item), np.integer):
            relidx = self.molsys.ix[self.mask][item]
            return Atom(self.molsys, relidx)

        if isinstance(item, slice):
            return AtomSelection(self.molsys, self.mask[item])

        elif isinstance(item, (np.ndarray, list, tuple)):
            item = np.asarray(item)
            if len(item) == 1 or item.sum() == 1:
                return Atom(self.molsys, self.mask[item])
            elif item.dtype == bool:
                item = np.argwhere(item).T[0]
                return AtomSelection(self.molsys, item)
            else:
                return AtomSelection(self.molsys, self.mask[item])

        elif hasattr(item, '__iter__'):
            if all([np.issubdtype(type(x), int) for x in item]):
                return AtomSelection(self.molsys, self.mask[item])

        raise TypeError('Only integer, slice type, and boolean mask arguments are supported at this time')

    def __len__(self):
        return len(self.mask)


class ResidueSelection(MolecularSystemBase):
    """
    An object containing a group of residues and all their atoms from a :class:`~MolSys` object. Note that if one atom of
    a residue is present in the AtomSelection used to create the group, all atoms of that residue will be present in
    the ResidueSelection object.

    .. note::
        Unlike a regular :class:`~AtomSelection` object, the ``resnames``, ``resnums``, ``segids`` and ``chains`` 
        attributes of this object are tracked with respect to residue not atom and changes made to these attributes
        on this object will not be present in the parent object and vice versa.

    Parameters
    ----------
    molsys : :class:`~MolSys`
        The :class:`~MolSys` object from which the AtomSelection belongs to.
    mask : np.ndarray
        An array of atom indices that defines the atoms of ``molsys`` that make up the atom selection.
    """

    def __init__(self, molsys, mask):

        resixs = np.unique(molsys.resixs[mask])
        self.mask = np.argwhere(np.isin(molsys.resixs, resixs)).flatten()
        self.molsys = molsys

        _, self.first_ix = np.unique(molsys.resixs[self.mask], return_index=True)

        self.resixs = self.resixs[self.first_ix].flatten()
        self.resnames = self.resnames[self.first_ix].flatten()
        self.resnums = self.resnums[self.first_ix].flatten()
        self.segids = self.segids[self.first_ix].flatten()
        self.chains = self.chains[self.first_ix].flatten()

    def __setattr__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, item):
        resixs = np.unique(self.molsys.resixs[self.mask])
        new_resixs = resixs[item]
        new_mask = np.argwhere(np.isin(self.molsys.resixs, new_resixs)).flatten()

        if np.issubdtype(type(item), int):
            return Residue(self.molsys, new_mask)
        elif isinstance(item, slice):
            return ResidueSelection(self.molsys, new_mask)
        elif hasattr(item, '__iter__'):
            if all([np.issubdtype(type(x), int) for x in item]):
                return ResidueSelection(self.molsys, new_mask)

        raise TypeError('Only integer and slice type arguments are supported at this time')

    def __len__(self):
        return len(self.first_ix)

    def __iter__(self):
        for resix, segid in zip(self.resixs, self.segids):
            mask = (self.molsys.resixs == resix) * (self.molsys.segids == segid)
            yield Residue(self.molsys, mask)


class SegmentSelection(MolecularSystemBase):
    """
    An object containing a group of segments and all their atoms from a :class:`~MolSys` object. Note that if one atom
    of a segment is present in the AtomSelection used to create the group, all atoms of that segment will be present in
    the SegmentSelection object.

    .. note::
        Unlike a regular :class:`~AtomSelection` object, the ``segids`` and ``chains`` 
        attributes of this object are tracked with respect to segements not atoms and changes made to these attributes
        on this object will not be present in the parent object and vice versa.

    Parameters
    ----------
    molsys : :class:`~MolSys`
        The :class:`~MolSys` object from which the AtomSelection belongs to.
    mask : np.ndarray
        An array of atom indices that defines the atoms of ``molsys`` that make up the atom selection.
    """

    def __init__(self, molsys, mask):
        self.seg_ixs = np.unique(molsys.segixs[mask])
        self.mask = np.argwhere(np.isin(molsys.segixs, self.seg_ixs)).flatten()
        self.molsys = molsys

        _, self.first_ix = np.unique(molsys.segixs[self.mask], return_index=True)

        self.segids = self.segids[self.first_ix].flatten()
        self.chains = self.chains[self.first_ix].flatten()

    def __setattr__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, item):
        segixs = np.unique(self.molsys.segixs[self.mask])
        new_segixs = segixs[item]
        new_mask = np.argwhere(np.isin(self.molsys.segixs, new_segixs))

        if np.issubdtype(type(item), int):
            return Segment(self.molsys, new_mask)
        elif isinstance(item, slice):
            return SegmentSelection(self.molsys, new_mask)
        elif hasattr(item, '__iter__'):
            if all([np.issubdtype(type(x), int) for x in item]):
                return SegmentSelection(self.molsys, new_mask)

        raise TypeError('Only integer and slice type arguments are supported at this time')

    def __len__(self):
        return len(self.first_ix)

    def __iter__(self):
        for new_segix in self.seg_ixs:
            new_mask = np.argwhere(np.isin(self.molsys.segixs, new_segix))
            yield Segment(self.molsys, new_mask)



class Atom(MolecularSystemBase):
    """
    An object containing information for a single atom.
    
    Parameters
    ----------
    molsys : :class:`~MolSys`
        The :class:`~MolSys` object from which the Atom belongs to.
    mask : np.ndarray
        The index that defines the atoms of ``molsys`` that make up the atom selection.
    """
    def __init__(self, molsys, mask):
        self.__dict__['molsys'] = molsys
        self.index = mask
        self.mask = mask

        self.name = molsys.names[self.index]
        self.altloc = molsys.altlocs[self.index]
        self.atype = molsys.types[self.index]
        self.resname = molsys.resnames[self.index]
        self.resnum = molsys.resnums[self.index]
        self.chain = molsys.segids[self.index]
        self.segid = molsys.chains[self.index]
        self.charge = molsys.charges[self.index]

        self.resn = self.resname
        self.resi = self.resnum
        self.altLoc = self.altloc
        self.type = self.atype
        self.resid = self.resnum


    @property
    def position(self):
        return self.molsys.coords[self.index]

    @property
    def bonds(self):
        idxs = self.topology.bonds_any_atom[self.mask]
        return [AtomSelection(self.molsys, idx) for idx in idxs]

    @property
    def angles(self):
        idxs = self.topology.angles_any_atom[self.mask]
        return [AtomSelection(self.molsys, idx) for idx in idxs]

    @property
    def dihedrals(self):
        idxs = self.topology.dihedrals_any_atom[self.mask]
        return [AtomSelection(self.molsys, idx) for idx in idxs]

    @property
    def bonded_atoms(self):
        all_idxs = np.unique(np.concatenate(self.topology.bonds_any_atom[self.mask]))
        all_idxs  = np.array([idx for idx in all_idxs if idx != self.mask])
        return AtomSelection(self.molsys, all_idxs)


class Residue(MolecularSystemBase):
    """
    An object representing a single residue.

    Parameters
    ----------
    molsys : :class:`~MolSys`
        The :class:`~MolSys` object from which the Residue belongs to.
    mask : int
        An array of atom indices that defines the atoms of the residue
    """

    def __init__(self, molsys, mask):
        resix = np.unique(molsys.resixs[mask])[0]
        self.mask = np.argwhere(np.isin(molsys.resixs, resix)).T[0]
        self.molsys = molsys

        self.resname = molsys.resnames[self.mask][0]
        self.resname = molsys.resnames[self.mask][0]

        self.resnum = molsys.resnums[self.mask][0]
        self.resid = self.resnum
        self.icode = molsys.icodes[self.mask][0]
        self.resindex = molsys.resixs[self.mask][0]
        self.segid = molsys.segids[self.mask][0]
        self.segindex = molsys.segixs[self.mask][0]
        self.chain = molsys.chains[self.mask][0]

    def __len__(self):
        return len(self.mask)

    def phi_selection(self):
        """
        Get an :class:`~AtomSelection` of the atoms defining the Phi backbone dihedral angle of the residue.

        Returns
        -------
        sel : AtomSelection
            An :class:`~AtomSelection` object containing the atoms that make up the Phi backbone dihedral angle of
            the selected residue.
        """

        prev = self.atoms.resixs[0] - 1

        maskN_CA_C = self.mask[np.isin(self.names, ['N', 'CA', 'C'])]
        maskC = np.argwhere(np.isin(self.molsys.resixs, prev)).flatten()
        maskC = maskC[self.molsys.names[maskC] == 'C']
        mask_phi = np.concatenate((maskC, maskN_CA_C))
        sel = self.molsys.atoms[mask_phi]
        return sel if len(sel) == 4 else None

    def psi_selection(self):
        """
        Get an :class:`~AtomSelection` of the atoms defining the Psi backbone dihedral angle of the residue.

        Returns
        -------
        sel : AtomSelection
            An :class:`~AtomSelection` object containing the atoms that make up the Psi backbone dihedral angle of
            the selected residue.
        """

        nex = self.atoms.resnums + 1
        nex = np.unique(nex[nex <= self.molsys.resnums.max()])

        maskN_CA_C = self.mask[np.isin(self.names, ['N', 'CA', 'C'])]
        maskN = np.argwhere(np.isin(self.molsys.resnums, nex)).flatten()
        maskN = maskN[self.molsys.names[maskN] == 'N']
        mask_psi = np.concatenate((maskN_CA_C, maskN))
        sel = self.molsys.atoms[mask_psi]
        return sel if len(sel) == 4 else None


class Segment(MolecularSystemBase):
    """
    An object representing a single segment of a molecular system.

    Parameters
    ----------
    molsys : MolSys
        The :class:`~MolSys` object from which the segment belongs to.
    mask : int
        An array of atom indices that defines the atoms of the segment.
    """

    def __init__(self, molsys, mask):
        segix = np.unique(molsys.segixs[mask])[0]
        self.mask = np.argwhere(np.isin(molsys.segixs, segix)).T[0]
        self.molsys = molsys

        self.segid = molsys.segids[segix]
        self.chain = molsys.chains[segix]


def byres(mask, molsys):
    """Advanced operator to select atoms by residue"""
    residues = np.unique(molsys.resixs[mask])
    mask = np.isin(molsys.resixs, residues)
    return mask


def unot(mask):
    """Unitary `not` operator"""
    return ~mask


def within(distance, mask, molsys):
    """
    Advanced logic operator to identify atoms within a user defined distance.

    Parameters
    ----------
    distance : float
        The distance window defining which atoms will be included in the selection.
    mask : np.ndarray
        A boolean array defining a subset of  atoms of a :class:`~MolSys` from which the distance cutoff will be
        measured by.
    molsys : :class:`~MolSys`
        A chiLife :class:`~MolSys` from which to select atoms from .

    Returns
    -------
    out_mask : np.ndarray
        A boolean array defining a subset of  atoms of a :class:`~MolSys` that are within the user defined distance
        of the user defined selection.
    """
    distance = float(distance)
    tree1 = cKDTree(molsys.coords)
    tree2 = cKDTree(molsys.molsys.coords[mask])
    results = np.concatenate(tree2.query_ball_tree(tree1, distance))
    results = np.unique(results)
    out_mask = np.zeros_like(mask, dtype=bool)
    out_mask[results] = True
    out_mask = out_mask * ~mask
    return out_mask


within.nargs = 1


def concat_molsys(systems):
    """
    Function to concatenate two or more :class:`~MolSys` objects into a single :class:`~MolSys` object. Atoms will be
    concatenated in the order they are placed in the list.

    Parameters
    ----------
    systems : List[MolSys]
        List of :class:`~MolSys` objects to concatenate.

    Returns
    -------
    mol : MolSys
        The concatenated `~MolSys` object.
    """
    record_types = []
    anames = []
    atypes = []
    resnames = []
    resnums = []
    resindices = []
    icodes = []
    trajectory = []
    segindices = []
    segids = []

    for sys in systems:

        resind = sys.resindices + resindices[-1][-1] if len(resindices) > 0 else sys.resindices
        record_types.append(sys.record_types)
        anames.append(sys.names)
        atypes.append(sys.atypes)
        resnames.append(sys.resnames)
        resnums.append(sys.resnums)
        icodes.append(sys.icodes)
        resindices.append(resind)
        trajectory.append(sys.positions)
        segindices.append(sys.segindices)
        segids.append([segid if len(segid) == 1 else 'A' for segid in sys.segids])

    record_types = np.concatenate(record_types)
    anames = np.concatenate(anames)
    atypes = np.concatenate(atypes)
    resnames = np.concatenate(resnames)
    resnums = np.concatenate(resnums)
    resindices = np.concatenate(resindices)
    icodes = np.concatenate(icodes)
    segindices = np.concatenate(segindices)
    segids = np.concatenate(segids)
    trajectory = np.concatenate(trajectory)

    mol = MolSys.from_arrays(anames, atypes, resnames, resindices, resnums, segindices,
                             segids=segids, trajectory=trajectory, record_types=record_types, icodes=icodes)

    return mol


import inspect
import re
import warnings
from copy import deepcopy
from functools import partial
from pathlib import Path
import numpy as np
from numpy.typing import ArrayLike
from itertools import combinations
from scipy.spatial import cKDTree
import igraph as ig
from scipy.stats import skewnorm, circstd
import scipy.optimize as opt
import MDAnalysis as mda


import chilife.io as io
import chilife.scoring as scoring

from .globals import SUPPORTED_RESIDUES, nataa_codes, dihedral_defs, ralt_prot_states
from .scoring import get_lj_rep, GAS_CONST
from .numba_utils import get_sasa as nu_getsasa
from .alignment_methods import alignment_methods, parse_backbone, local_mx, global_mx
from .protein_utils import FreeAtom, guess_mobile_dihedrals, get_dihedral, get_angle
from .pdb_utils import get_bb_candidates, get_backbone_atoms
from .MolSys import MolSys, MolecularSystemBase
from .MolSysIC import MolSysIC

default_energy_func = scoring.ljEnergyFunc()

class RotamerEnsemble:
    """Create new RotamerEnsemble object.

    Parameters
    ----------
    res : string
        3-character name of desired residue, e.g. R1A.
    site : int
        MolSys residue number to attach library to.
    protein : MDAnalysis.Universe, MDAnalysis.AtomGroup
        Object containing all protein information (coords, atom types, etc.)
    chain : str
        MolSys chain identifier to attach spin label to.
    rotlib : str
        Rotamer library to use for constructing the RotamerEnsemble
    **kwargs : dict
        minimize: bool
            Switch to turn on/off minimization. During minimization each rotamer is optimized in dihedral space
            with respect to both internal and external clashes, and deviation from the starting conformation in
            dihedral space.
        min_method: str
            Name of the minimization algorithm to use. All ``scipy.optimize.minimize`` algorithms are available
            and include: ‘Nelder-Mead’, ‘Powell’, ‘CG’, ‘BFGS’, ‘Newton-CG’, ‘L-BFGS-B’, ‘TNC’, ‘COBYLA’, ‘SLSQP’,
            ‘trust-constr’, ‘dogleg’, ‘trust-ncg’, ‘trust-exact’, ‘trust-krylov’, and custom.
        exclude_nb_interactions: int:
            When calculating internal clashes, ignore 1-``exclude_nb_interactions`` interactions and below. Defaults
            to ignore 1-3 interactions, i.e. atoms that are connected by 2 bonds or fewer will not have a steric
            effect on each other.
        eval_clash : bool
            Switch to turn clash evaluation on (True) and off (False).
        forcefield: str
            Name of the forcefield you wish to use to parameterize atoms for the energy function. Currently, supports
            `charmm` and `uff`
        energy_func : callable
            Python function or callable object that takes a protein or a RotamerEnsemble object as input and
            returns an energy value (kcal/mol) for each atom of each rotamer in the ensemble. See also
            :mod:`Scoring <chiLife.scoring>` . Defaults to a capped lennard-jones repulsive energy function.
        temp : float
           Temperature to use when running ``energy_func``. Only used if ``energy_func`` accepts a temperature
           argument  Defaults to 298 K.
        clash_radius : float
           Cutoff distance (angstroms) for inclusion of atoms in clash evaluations. This distance is measured from
           ``clash_ori`` Defaults to the longest distance between any two atoms in the rotamer ensemble plus 5
           angstroms.
        clash_ori : str
           Atom selection to use as the origin when finding atoms within the ``clash_radius``. Defaults to 'cen',
           the centroid of the rotamer ensemble heavy atoms.
        ignore_wateres : bool
            Determines whether water molecules (resnames HOH and OH2) are ignored during clash evaluation.
            Defaults to True.
        protein_tree : Scipy.spatial.cKDTree
           KDTree of atom positions for fast distance calculations and neighbor detection. Defaults to None
        trim: bool
            When true, the lowest `trim_tol` fraction of rotamers in the ensemble will be removed.
        trim_tol: float
            Tolerance for trimming rotamers from the ensemble. trim_tol=0.005 means the bottom 0.5% of rotamers
            will be removed.
        alignment_method : str
           Method to use when attaching or aligning the rotamer ensemble backbone with the protein backbone.
           Defaults to ``'bisect'`` which aligns the CA atom, the vectors bisecting the N-CA-C angle and the
           N-CA-C plane.
        sample : int, bool
           Argument to use the off-rotamer sampling method. If ``False`` or ``0`` the off-rotamer sampling method
           will not be used. If ``int`` the ensemble will be generated with that many off-rotamer samples.
        dihedral_sigmas : float, numpy.ndarray
           Standard deviations of dihedral angles (degrees) for off rotamer sampling. Can be a single number for
           isotropic sampling, a vector to define each dihedral individually or a matrix to define a value for
           each rotamer and each dihedral. Setting this value to np.inf will force uniform (accessible volume)
           sampling. Defaults to 35 degrees.
        weighted_sampling : bool
           Determines whether the rotamer ensemble is sampled uniformly or based off of their intrinsic weights.
           Defaults to False.
        use_H : bool
           Determines if hydrogen atoms are used or not. Defaults to False.
       """


    def __init__(self, res, site=None, protein=None, chain=None, rotlib=None, **kwargs):


        self.res = res
        if site is None and protein is not None:
            raise ValueError('A protein has been provided but a site has not. If you wish to construct an ensemble '
                             'associated with a protein you must include the site you wish to model.')
        elif site is None:
            site = 1
            icode = ""
        elif isinstance(site, str):
            site_split = re.split(r'(\D+)',site)
            site = site_split[0]
            icode = site_split[1] if len(site_split) > 1 else ""
        else:
            icode = ""


        self.site = site
        self.icode = icode
        self.resnum = self.site
        self.chain = chain if chain is not None else guess_chain(protein, self.site)
        self.selstr = f"resid {self.site} and segid {self.chain} and not altloc B"
        self.nataa = ""
        self.input_kwargs = kwargs
        self.__dict__.update(assign_defaults(kwargs))
        self.input_kwargs.pop('energy_func', None)
        self.protein = protein

        # Convert string arguments for alignment_method to respective function
        if isinstance(self.alignment_method, str):
            self.alignment_method = alignment_methods[self.alignment_method]

        lib = self.get_lib(rotlib)
        self.__dict__.update(lib)
        self._weights = self.weights / self.weights.sum()

        if len(self.sigmas) != 0 and 'dihedral_sigmas' not in kwargs:
            self.sigmas[self.sigmas == 0] = self.dihedral_sigmas
        else:
            self.set_dihedral_sampling_sigmas(self.dihedral_sigmas)

        self._rsigmas = np.deg2rad(self.sigmas)
        with np.errstate(divide='ignore'):
            self._rkappas = 1 / self._rsigmas**2

        self._rkappas = np.clip(self._rkappas, 0, 1e16)

        # Remove hydrogen atoms unless otherwise specified
        if not self.use_H:
            self.H_mask = self.atom_types != "H"
            self._coords = self._coords[:, self.H_mask]
            self.atom_types = self.atom_types[self.H_mask]
            self.atom_names = self.atom_names[self.H_mask]

        self.ic_mask = np.argwhere(np.isin(self.internal_coords.atom_names, self.atom_names)).flatten()
        self._lib_coords = self._coords.copy()
        self._lib_dihedrals = self._dihedrals.copy()
        self._lib_IC = self.internal_coords

        if self.clash_radius is None:
            self.clash_radius = np.linalg.norm(self.clash_ori - self.coords, axis=-1).max() + 5

        # Parse important indices
        self.aln_idx = np.squeeze(np.argwhere(np.isin(self.atom_names, self.aln_atoms)))
        self.backbone_idx = np.squeeze(np.argwhere(np.isin(self.atom_names, self.backbone_atoms)))
        self.side_chain_idx = np.argwhere(np.isin(self.atom_names, self.backbone_atoms, invert=True)).flatten()

        self._graph = ig.Graph(edges=self.bonds)
        self.aidx, self.bidx = [list(x) for x in zip(*self.non_bonded)]

        # Allocate variables for clash evaluations
        self.atom_energies = None
        self.clash_ignore_idx = None
        self.partition = 1

        # Assign a name to the label
        self.name = self.rotlib
        if self.site is not None:
            self.name = self.nataa + str(self.site) + self.rotlib
        if self.chain not in ('A', None):
            self.name += f"_{self.chain}"

        self.update(no_lib=True)

        # Store atom information as atom objects
        self.atoms = [
            FreeAtom(name, atype, idx, self.res, self.site, coord)
            for idx, (coord, atype, name) in enumerate(
                zip(self._coords[0], self.atom_types, self.atom_names)
            )
        ]

    def __str__(self):
        return (
            f"Rotamer ensemble with {np.size(self.weights)} members\n" +
            f"  Name: {self.name}\n" +
            f"  Label: {self.res}\n" +
            f"  Site: {self.site}\n" +
            f"  Dihedral definitions:\n" +
            f"\n".join([f'    {d}' for d in self.dihedral_atoms]) +
            f"\n"
        )

    def __repr__(self):
        return str(self)

    @classmethod
    def from_pdb(
        cls,
        pdb_file,
        res=None,
        site=None,
        protein=None,
        chain=None,
    ):
        """

        Parameters
        ----------
        pdb_file :
            
        res :
             (Default value = None)
        site :
             (Default value = None)
        protein :
             (Default value = None)
        chain :
             (Default value = None)

        Returns
        -------

        """
        # TODO: create pdb reader and implement from_pdb method
        raise NotImplementedError('Creation of a RotamerEnsemble from a PDB file is not yet implemented.')

        return cls(res, site, protein, chain)

    @classmethod
    def from_mda(cls, residue, **kwargs):
        """Create RotamerEnsemble from the MDAnalysis.Residue for chiLife.Residue

        Parameters
        ----------
        residue : MDAnalysis.Residue, chiLife.Residue
            A Residue object from which to create the RotamerEnsemble.
        **kwargs : dict
            Additional keyword arguments to use for creating the RotamerEnsemble.

        Returns
        -------
        cls : RotamerEnsemble
            A RotamerEnsemble object created from the provided Residue object.
        """

        res = residue.resname


        if kwargs.get('use_H', False) and res in ralt_prot_states:
            if res != 'HIS':
                res = ralt_prot_states[res].get(len(residue.atoms), res)
            elif len(residue.atoms) == 18:
                res = 'HIP'
            elif len([x for x in residue.atoms.names if 'HE' in x]) == 2:
                res = 'HIE'
            elif len([x for x in residue.atoms.names if 'HD' in x]) == 2:
                res = 'HID'
            else:
                res = 'HIS'


        site = residue.resid
        chain = residue.segid
        protein = residue.universe
        return cls(res, site, protein, chain, **kwargs)

    @classmethod
    def from_trajectory(cls, traj, site, chain=None, energy=None, burn_in=0, **kwargs):
        """
        Create a RotamerEnsemble object from a trajectory.

        Parameters
        ----------
        traj : MDAnalysis.Universe, MDAnalysis.AtomGroup, chiLife.MolecularSystemBase
            The trajectory over which the RotamerEnsemble is to be created.
        site : res
            Residue number of the residue to create a rotamer ensemble for.
        chain : str
            Chain identifier of the residue to create a RotamerEnsemble for.
        energy : ArrayLike
            Potential energy of the residue at each frame of the trajectory in kcal/mol
        burn_in : int
             Number of frames to skip from the trajectory. Used to skip burn-in period when performing MCMC sampling.
        **kwargs : dict
            Additional keyword arguments to use for creating the RotamerEnsemble.

            dihedral_atoms : list
                A list of atoms defining the mobile dihedral angles of the residue. Primarily used when creating an
                ensemble of a residue type chiLife is unaware of (see :ref:`mobile_dihedrals`
            temp : float
                Temperature to use when evaluating conformational populations from user supplied ``energy`` argument.
            spin_atoms : list, dict
                A list or dict defining the spin atoms of the label and their relative populations (see
                :ref:`spin_atoms`).

        Returns
        -------
        cls : RotamerEnsemble
            A RotamerEnsemble object created from the provided trajectory.
        """
        if burn_in >= len(traj.universe.trajectory):
            raise ValueError("Burn in is longer than the provided trajectory.")

        chain = guess_chain(traj, site) if chain is None else chain

        if isinstance(traj, (mda.AtomGroup, mda.Universe)):
            if not hasattr(traj.universe._topology, "altLocs"):
                traj.add_TopologyAttr('altLocs', ["A"] * len(traj.atoms))

        res = traj.select_atoms(f"segid {chain} and resid {site} and not altloc B")

        resname = res.residues[0].resname
        ddefs = kwargs.pop('dihedral_atoms', dihedral_defs.get(resname, ()))

        traj = traj.universe if isinstance(traj, mda.AtomGroup) else traj
        coords = np.array([res.atoms.positions for ts in traj.trajectory[burn_in:]])

        _, unique_idx, non_unique_idx = np.unique(coords, axis=0, return_inverse=True, return_index=True)
        coords = coords[unique_idx]

        if energy is not None:
            energy = energy[burn_in:]  # - energy[burn_in]
            energy = np.array([energy[non_unique_idx == idx].mean() for idx in range(len(unique_idx))])
            T = kwargs.setdefault("temp", 298)
            pi = np.exp(-energy / (scoring.GAS_CONST * T))
        else:
            pi = np.ones(len(traj.trajectory[burn_in:]))
            pi = np.array([pi[non_unique_idx == idx].sum() for idx in range(len(unique_idx))])

        pi /= pi.sum()
        weights = pi

        if not kwargs.setdefault('use_H', True):
            res = res.select_atoms('not type H')

        res = MolSys.from_atomsel(res, frames = unique_idx)

        ICs = MolSysIC.from_atoms(res)
        ICs.shift_resnum(-(site - 1))

        if ddefs == ():
            ddefs = guess_mobile_dihedrals(ICs)

        dihedrals = np.array([ic.get_dihedral(1, ddefs) for ic in ICs])
        sigmas = kwargs.get('sigmas', np.array([]))

        bb_candidates = get_bb_candidates(res.names, resname)
        aln_atoms = kwargs.get('ln_atoms', [])
        if not bb_candidates and not aln_atoms:
            raise RuntimeError('No suitable backbone candidates were found. Please use from_trajectory on residues with'
                               'standard backbone atom names or specify alignment backbone atoms with the aln_atoms '
                               'keyword argument.')
        elif aln_atoms:
            graph = res.topology.graph
            root_idx = np.argwhere(res.names == aln_atoms[1]).flat[0]
            aname_lst = ICs.atom_names.tolist()
            neighbor_idx = [aname_lst.index(a) for a in aln_atoms[::2]]
            backbone_atoms = get_backbone_atoms(graph, root_idx, neighbor_idx)

        else:
            backbone_atoms = [b for b in bb_candidates if b in res.names]
            aln_atoms = ['N', 'CA', 'C'] if np.all(np.isin(['N', 'CA', 'C'], backbone_atoms)) else ["O4'", "C1'", "C2'"]

        lib = {'rotlib': f'{resname}_from_traj',
               'resname': resname,
               'coords': np.atleast_3d(coords),
               'internal_coords': ICs,
               'weights': weights,
               'atom_types': res.types.copy(),
               'atom_names': res.names.copy(),
               'dihedral_atoms': ddefs,
               'dihedrals': np.rad2deg(dihedrals),
               'aln_atoms': aln_atoms,
               'backbone_atoms': backbone_atoms,
               '_dihedrals': dihedrals.copy(),
               '_rdihedrals': dihedrals,
               'sigmas': sigmas,
               '_rsigmas': np.deg2rad(sigmas),
               'type': 'chilife rotamer library from a trajectory',
               'format_version': 1.0}

        if 'spin_atoms' in kwargs:
            spin_atoms = kwargs.pop('spin_atoms')
            if isinstance(spin_atoms, (list, tuple, np.ndarray)):
                spin_atoms = {sa: 1 / len(spin_atoms) for sa in spin_atoms}
            elif isinstance(spin_atoms, dict):
                pass
            else:
                raise TypeError('the `spin_atoms` kwarg must be a dict, list or tuple')

            lib['spin_atoms'] = np.array(list(spin_atoms.keys()))
            lib['spin_weights'] = np.array(list(spin_atoms.values()))

        elif lib_path := io.get_possible_rotlibs(resname, suffix='rotlib', extension='.npz'):
            with np.load(lib_path) as f:
                if 'spin_atoms' in f:
                    lib['spin_atoms'] = f['spin_atoms']
                    lib['spin_weights'] = f['spin_weights']

        kwargs.setdefault('eval_clash', False)
        kwargs.setdefault('_match_backbone', False)
        return cls(resname, site, chain=chain, rotlib=lib, **kwargs)

    def update(self, no_lib=False):
        # Sample from library if requested
        if self._sample_size and len(self.dihedral_atoms) > 0:
            mask = self.sigmas.sum(axis=0) == 0

            # Draw samples
            coords, weights, internal_coords = self.sample(self._sample_size, off_rotamer=~mask, return_dihedrals=True)

            # Remove structures with internal clashes
            dist = np.linalg.norm(coords[:, self.aidx] - coords[:, self.bidx], axis=2)
            sidx = np.atleast_1d(np.squeeze(np.argwhere(np.all(dist > 2, axis=1))))
            self.internal_coords = internal_coords
            self.internal_coords.use_frames(sidx)
            dihedrals = np.asarray(
                [self.internal_coords.get_dihedral(1, self.dihedral_atoms) for ts in self.internal_coords.trajectory]
            )
            self._dihedrals = np.rad2deg(dihedrals)
            self._coords, self.weights = coords[sidx], weights[sidx]

        elif not no_lib:
            # Reset to library
            self._coords = self._lib_coords.copy()
            self._dihedrals = self._lib_dihedrals.copy()
            self.internal_coords = self._lib_IC
            self.weights = self._weights.copy()

        if self.protein is not None:
            self.protein_setup()
        else:
            self.ICs_to_site(self.origin, self.mx)

    def protein_setup(self):
        self.to_site()

        # Get weight of current or closest rotamer
        clash_ignore_idx = self.protein.select_atoms(f"resid {self.site} and segid {self.chain}").ix
        self.clash_ignore_idx = np.argwhere(np.isin(self.protein.ix, clash_ignore_idx)).flatten()
        protein_clash_idx = self.protein_tree.query_ball_point(self.clash_ori, self.clash_radius)
        self.protein_clash_idx = [idx for idx in protein_clash_idx if idx not in self.clash_ignore_idx]

        if hasattr(self.energy_func, 'prepare_system'):
            self.energy_func.prepare_system(self)

        if self._coords.shape[1] == len(self.clash_ignore_idx):
            RMSDs = np.linalg.norm(
                self._coords - self.protein.atoms[self.clash_ignore_idx].positions[None, :, :],
                axis=(1, 2)
            )

            idx = np.argmin(RMSDs)
            self.current_weight = self.weights[idx]
        else:
            self.current_weight = 0

        # Evaluate external clash energies and reweigh rotamers
        if self._minimize and self.eval_clash:
            raise RuntimeError('Both `minimize` and `eval_clash` options have been selected, but they are incompatible.'
                               'Please select only one. Also note that minimize performs its own clash evaluations so '
                               'eval_clash is not necessary.')
        elif self.eval_clash:
            self.evaluate()

        elif self._minimize:
            self.minimize()

    def to_rotlib(self,
                  libname: str = None,
                  description: str = None,
                  comment: str = None,
                  reference: str = None) -> None:
        """
        Saves the current RotamerEnsemble as a rotamer library file.

        Parameters
        ----------
        libname : str
            Name of the rotamer library.
        description : str
            Description of the rotamer library.
        comment : str
            Additional comments about the rotamer library.
        reference : str
            References associated with the rotamer library.
        """
        if libname is None:
            libname = self.name.lstrip('0123456789.- ')

        if description is None:
            description = (f'Rotamer library made with chiLife using `to_rotlib` method'
                           f'of a rotamer ensemble.')
        ICs = self.internal_coords.copy()

        # Remove chain operators to align all labels on backbone
        # ICs.chain_operators = None
        coords = ICs.protein.trajectory.coordinate_array
        bb_idx = np.argwhere(np.isin(ICs.atom_names, ['N', 'CA', 'C'])).flatten()
        for i in range(len(coords)):
            ori, mx = local_mx(*coords[i, bb_idx])
            coords[i] = (coords[i] - ori) @ mx

        lib = {'rotlib': libname,
               'resname': self.res,
               'coords': coords,
               'internal_coords': ICs,
               'weights': self.weights,
               'atom_types': self.atom_types.copy(),
               'atom_names': self.atom_names.copy(),
               'dihedral_atoms': self.dihedral_atoms,
               'dihedrals': self.dihedrals,
               'aln_atoms': self.aln_atoms,
               'backbone_atoms': self.backbone_atoms,
               'sigmas': self.sigmas,
               'type': 'chilife rotamer library',
               'description': description,
               'comment': comment,
               'reference': reference,
               'format_version': 1.4}

        if hasattr(self, 'spin_atoms'):
            lib['spin_atoms'] = self.spin_atoms
            lib['spin_weights'] = self.spin_weights

        np.savez(Path().cwd() / f'{libname}_rotlib.npz', **lib, allow_pickle=True)

    def __eq__(self, other):
        """Equivalence measurement between a spin label and another object. A SpinLabel cannot be equivalent to a
        non-SpinLabel Two SpinLabels are considered equivalent if they have the same coordinates and weights
        TODO: Future consideration -- different orders (argsort weights)"""

        if not isinstance(other, RotamerEnsemble):
            return False
        elif self._coords.shape != other._coords.shape:
            return False

        return np.all(np.isclose(self._coords, other._coords)) and np.all(
            np.isclose(self.weights, other.weights)
        )

    def set_site(self, site):
        """Assign RotamerLibrary to a site.

        Parameters
        ----------
        site : int
            Residue number to assign SpinLabel to.
        """
        self.site = site
        self.to_site()

    def copy(self, site: int = None, chain: str = None, rotlib: dict = None):
        """Create a deep copy of the spin label. Assign new site, chain or rotlib  if desired. Useful when labeling
        homo-oligomers or proteins with repeating units/domains.

        Parameters
        ----------
        site : int
             New site number for the copy.
        chain : str
             New chain identifier for the copy.
        rotlib : dict
            New (base) rotamer library for the dict. Used primarily when copying dRotamerEnsembels.

        Returns
        -------
        new_copy : chilife.RotamerEnsemble
            A deep copy of the original RotamerLibrary
        """
        new_copy = self._base_copy(self._rotlib)
        for item in self.__dict__:
            if isinstance(item, np.ndarray) or item == 'internal_coords':
                new_copy.__dict__[item] = self.__dict__[item].copy()
            elif item not in  ("_protein", '_lib_IC', '_rotlib'):
                new_copy.__dict__[item] = deepcopy(self.__dict__[item])
            elif self.__dict__[item] is None:
                new_copy.protein = None
            else:
                new_copy.__dict__[item] = self.__dict__[item]

        if site is not None:
            new_copy.site = site
        if chain is not None:
            new_copy.chain = chain
        return new_copy

    def _base_copy(self, rotlib=None):
        return RotamerEnsemble(self.res, self.site, rotlib=rotlib, chain=self.chain)

    def to_site(self, site_pos: ArrayLike = None) -> None:
        """Move spin label to new site

        Parameters
        ----------
        site_pos : ArrayLike
            3x3 array of ordered backbone atom coordinates of new site (N CA C) (Default value = None)
        """
        if site_pos is None:
            bbs = parse_backbone(self, kind="global")
        else:
            bbs = site_pos

        if len(bbs) < 3 :
            raise RuntimeError(f'The residue/rotlib you have selected {self.res}, does not share a compatible backbone '
                               f'with the residue you are trying to label. Check the site and rotlib and try again.\n'
                               f'The label backbone atoms: {self.backbone_atoms}')

        mx, ori = global_mx(*bbs, method=self.alignment_method)

        # if self.alignment_method not in {'fit', 'rosetta'}:
        bbs = parse_backbone(self, kind="local")
        old_ori, ori_mx = local_mx(*bbs, method=self.alignment_method)
        self._coords -= old_ori
        cmx = ori_mx @ mx

        self._coords = np.einsum("ijk,kl->ijl", self._coords, cmx) + ori
        self.ICs_to_site(ori, mx)
        if self._match_backbone:
            self.backbone_to_site()

    def ICs_to_site(self, cori, cmx):
        """ Modify the internal coordinates to be aligned with the site that the RotamerEnsemble is attached to"""

        # Update chain operators
        ic_backbone = np.squeeze(self.internal_coords.coords[:3])

        if self.alignment_method.__name__ == 'fit_alignment':
            N, CA, C = parse_backbone(self, kind="global")
            ic_backbone = np.array([[ic_backbone[0], N[1]],
                                    [ic_backbone[1], CA[1]],
                                    [ic_backbone[2], C[1]]])

        self.ic_ori, self.ic_mx = local_mx(*ic_backbone, method=self.alignment_method)
        m2m3 = self.ic_mx @ self.mx
        op = {}

        new_mx = self.internal_coords.chain_operators[0]["mx"] @ m2m3
        new_ori = (self.internal_coords.chain_operators[0]["ori"] - self.ic_ori) @ m2m3 + self.origin
        op[0] = {"mx": new_mx, "ori": new_ori}

        self.internal_coords.chain_operators = [op]

        # Update backbone conf
        alist = ["O"] if not self.use_H else ["H", 'O']
        for atom in alist:
            mask = self.internal_coords.atom_names == atom
            if any(mask) and self.protein is not None:
                idx = np.argwhere(self.internal_coords.atom_names == atom).flat[0]
                dihe_def = self.internal_coords.z_matrix_names[idx]
                p = self.protein.select_atoms(
                    f"segid {self.chain} and resid {self.site} and name {' '.join(dihe_def)} and not altloc B"
                ).positions

                if len(p) > 4:
                    # Guess that the closest H to the nitrogen is the H we are looking for
                    HN_idx = np.argmin(np.linalg.norm(p[3:] - p[0], axis=1)) + 3
                    p = p[[0, 1, 2, HN_idx]]
                elif len(p) == 3:
                    continue

                if atom == 'H':
                    # Reorder
                    p = p[[2, 1, 0, 3]]

                dihe = get_dihedral(p)
                ang = get_angle(p[1:])
                bond = np.linalg.norm(p[-1] - p[-2])
                idx = np.squeeze(np.argwhere(mask))
                chain = self.internal_coords.chains[0]
                resnum = self.internal_coords.atoms[0].resid
                if atom == 'O' and (tag := (chain, resnum, 'C', 'CA')) in self.internal_coords.chain_res_name_map:
                    additional_idxs = self.internal_coords.chain_res_name_map[tag]

                delta = self.internal_coords.trajectory.coordinate_array[:, idx, 2] - dihe
                self.internal_coords.trajectory.coordinate_array[:, idx] = bond, ang, dihe
                if atom == "O" and 'additional_idxs' in locals():
                    self.internal_coords.trajectory.coordinate_array[:, additional_idxs, 2] -= delta[:, None]

    def backbone_to_site(self):
        """Modify backbone atoms to match the site that the RotamerEnsemble is attached to """
        # Keep protein backbone dihedrals for oxygen and hydrogen atoms

        for atom in self.backbone_atoms:

            if atom in self.aln_atoms: continue  # TODO update test to change this
            mask = self.atom_names == atom
            if any(mask) and self.protein is not None:

                pos = self.protein.select_atoms(
                    f"segid {self.chain} and resid {self.site} "
                    f"and name {atom} and not altloc B"
                ).positions

                if pos.shape == (1, 3):
                    self._coords[:, mask] = np.squeeze(pos)
                elif pos.shape == (3,):
                    self._coords[:, mask] = np.squeeze(pos)
                elif pos.shape[0] > 1:
                    idx = np.argmin(np.linalg.norm(pos - self.backbone[0], axis=1))
                    self._coords[:, mask] = np.squeeze(pos[idx])
                else:
                    pass

    def sample(self, n=1, off_rotamer=False, **kwargs):
        """Randomly sample a rotamer in the library.

        Parameters
        ----------
        n : int
             number of rotamers to sample.
        off_rotamer : bool | List[bool]
            If False the rotamer library used to construct the RotamerEnsemble will be sampled using the exact
            dihedrals. If True, all mobile dihedrals will undergo minor perturbations. A list indicates a per-dihedral
            switch to turn on/off off-rotamer sampling of a particular dihedrals, corresponding to `self.dihedral_atoms`
            e.g. `off_+rotamer = [False, False, False, True, True]` for R1 will only sample off-rotamers for χ4 and χ5.
        **kwargs : dict
            return_dihedrals : bool
                If True, sample will return a MolSysIC object of the sampled rotamer

        Returns
        -------
        coords : ArrayLike
            The 3D coordinates of the sampled rotamer(s)
        new_weights : ArrayLike
            New weights (relative) of the sampled rotamer(s)
        ICs : List[chilife.MolSysIC] (Optional)
            Internal coordinate (MolSysIC) objects of the rotamer(s).
        """

        if not self.weighted_sampling:
            idx = np.random.randint(len(self._weights), size=n)
        else:
            idx = np.random.choice(len(self._weights), size=n, p=self._weights)

        if not hasattr(off_rotamer, "__len__"):
            off_rotamer = (
                [off_rotamer] * len(self.dihedral_atoms)
                if len(self.dihedral_atoms) > 0
                else [True]
            )
        else:
            off_rotamer = off_rotamer[: len(self.dihedral_atoms)]

        if not any(off_rotamer):
            if not np.allclose(np.squeeze(self._lib_coords[0, self.backbone_idx]), self.backbone):

                #####Should not be computing this every time
                N, CA, C = self.aln
                mx, ori = global_mx(N, CA, C, method=self.alignment_method)
                N, CA, C = np.squeeze(self._lib_coords[0, self.aln_idx])
                old_ori, ori_mx = local_mx(N, CA, C, method=self.alignment_method)
                ######

                self._lib_coords -= old_ori
                mx = ori_mx @ mx

                self._lib_coords = np.einsum("ijk,kl->ijl", self._lib_coords, mx) + ori
                for atom in ["H", "O"]:
                    mask = self.atom_names == atom
                    if any(mask) and self.protein is not None:
                        pos = self.protein.select_atoms(
                            f"segid {self.chain} and resid {self.site} "
                            f"and name {atom} and not altloc B"
                        ).positions
                        if len(pos) > 0:
                            self._lib_coords[:, mask] = pos[0] if len(pos.shape) > 1 else pos

            return np.squeeze(self._lib_coords[idx]), np.squeeze(self._weights[idx])

        if len(self.dihedral_atoms) == 0:
            return self._coords, self.weights, self.internal_coords
        elif hasattr(self, "internal_coords"):
            returnables = self._off_rotamer_sample(idx, off_rotamer, **kwargs)
            return [np.squeeze(x) if isinstance(x, np.ndarray) else x for x in returnables]

        else:
            raise AttributeError(
                f"{type(self)} objects require both internal coordinates ({type(self)}.internal_coords "
                f"attribute) and dihedral standard deviations ({type(self)}.sigmas attribute) to "
                f"perform off rotamer sampling"
            )

    def _off_rotamer_sample(self, idx, off_rotamer, **kwargs):
        """Perform off rotamer sampling. Primarily a helper function for `RotamerEnsemble.sample()`

        Parameters
        ----------
        idx : int
             Index of the rotamer of the rotamer library that will be perturbed.
        off_rotamer : List[bool]
             A list indicates a per-dihedral bools indicating which dihedral angles of `self.dihedral_atoms` will
             have off rotamer sampling. e.g. `off_+rotamer = [False, False, False, True, True]` for R1 will only sample
             off-rotamers for χ4 and χ5.
        **kwargs : dict
            return_dihedrals : bool
                If True, sample will return a MolSysIC object of the sampled rotamer

        Returns
        -------
        coords : ArrayLike
            The 3D coordinates of the sampled rotamer(s)
        new_weights : ArrayLike
            New weights (relative) of the sampled rotamer(s)
        ICs : List[chilife.MolSysIC] (Optional)
            Internal coordinate (MolSysIC) objects of the rotamer(s).
        """

        # Use accessible volume sampling if only provided a single rotamer
        if len(self._weights) == 1 or np.all(np.isinf(self.sigmas)):
            new_dihedrals = np.random.random((len(idx), len(off_rotamer))) * 2 * np.pi
            new_weights = np.ones(len(idx))

        #  sample from von mises near rotamer unless more information is provided
        elif self.skews is None:
            new_dihedrals = np.random.vonmises(self._rdihedrals[idx][:, off_rotamer],
                                               self._rkappas[idx][:, off_rotamer])
            new_weights = np.ones(len(idx))
        # Sample from skewednorm if skews are provided
        # else:
        #     deltas = skewnorm.rvs(a=self.skews[idx], loc=self.locs[idx], scale=self.sigmas[idx])
        #     pdf = skewnorm.pdf(deltas, a=self.skews[idx], loc=self.locs[idx], scale=self.sigmas[idx])
        #     new_dihedrals = np.deg2rad(self.dihedrals[idx] + deltas)
        #     new_weight = pdf.prod()

        else:
            new_weights = np.ones(len(idx))

        new_weights = self._weights[idx] * new_weights

        z_matrix = self._lib_IC.batch_set_dihedrals(idx, new_dihedrals, 1, self.dihedral_atoms[off_rotamer])
        ICs = self.internal_coords.copy()
        ICs.load_new(z_matrix)
        coords = ICs.protein.trajectory.coordinate_array[:, self.ic_mask]

        if kwargs.setdefault("return_dihedrals", False):
            return coords, new_weights, ICs
        else:
            return coords, new_weights

    def update_weight(self, weight: float) -> None:
        """
         Function to assign `self.current_weight`, which is the estimated weight of the rotamer currently occupying the
         attachment site on `self.protein`. This is only relevant if the residue type on `self.protein` is the same as
         the  RotamerLibrary.

        Parameters
        ----------
        weight : float
            New weight for the current residue.
        """
        self.current_weight = weight

    def minimize(self, callback=None):
        """
        Minimize the energy of the rotamers in dihedral space and re-weight with respect to the new energies.

        Parameters
        ----------
        callback : callable
            A callable function to be passed as the ``scipy.optimize.minimize`` function. See the
            `scipy documentation <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_
            for details.
        """
        dummy = self.copy()

        scores = np.array([self._min_one(i, ic, dummy, callback=callback) for i, ic in enumerate(self.internal_coords)])
        scores -= scores.min()

        self.weights *= np.exp(-scores / (GAS_CONST * self.temp) / np.exp(-scores).sum())
        self.weights /= self.weights.sum()
        if self._do_trim:
            self.trim_rotamers()

    def _objective(self, dihedrals, ic1, dummy):
        """
        Objective function for the minimization procedure.

        Parameters
        ----------
        dihedrals : ArrayLike
            Array of dihedral angles to generate a score for.
        ic1 : chilife.MolSysIC
            Internal coordinate objects
        dummy : chilife.RotamerEnsemble
            A dummy ensemble to use for applying dihedrals to and testing the score so that the parent rotamer ensemble
            does not need to be modified.

        Returns
        -------
        energy : float
            The "energy" score of the rotamer with the provided dihedrals.
        """

        ic1.set_dihedral(dihedrals[: len(self.dihedral_atoms)], 1, self.dihedral_atoms)
        coords = ic1.to_cartesian()[self.ic_mask]
        dummy._coords = np.atleast_3d([coords[self.ic_mask]])
        r = np.linalg.norm(coords[self.aidx] - coords[self.bidx], axis=1)

        # Faster to compute lj here
        lj = self.irmin_ij / r
        lj = lj * lj *lj
        lj = lj * lj

        # attractive forces are needed, otherwise this term will perpetually push atoms apart
        internal_energy = self.ieps_ij * (lj * lj - 2 * lj)
        external_energy = self.energy_func(dummy)
        energy = external_energy.sum() + internal_energy.sum()
        return energy

    def _min_one(self, i, ic, dummy, callback=None):
        """
        Perform a single minimization on a member of the underlying rotamer library in dihedral space.

        Parameters
        ----------
        i : int
            index of the rotamer (of the underlying library) to be minimized.
        ic : chilife.MolSysIC
            Internal coordinate object of the rotamer to be minimized.
        dummy : chilife.RotamerEnsemble
            A dummy ensemble to perform manipulations on while minimizing.

        Returns
        -------
        energy : float
            The final minimized "energy" of the objective function plus a modifier based on the deviations from the
            parent rotamer in the rotamer library.
        """
        if callback is not None:
            if 'i' in inspect.signature(callback).parameters:
                callback = partial(callback, i=i)

        d0 = ic.get_dihedral(1, self.dihedral_atoms)

        lb = d0 - np.deg2rad(40)
        ub = d0 + np.deg2rad(40)  #
        bounds = np.c_[lb, ub]
        xopt = opt.minimize(self._objective, x0=d0, args=(ic, dummy),
                            bounds=bounds, method=self.min_method,
                            callback=callback)
        self._coords[i] = ic.coords[self.ic_mask]
        tors = d0 - xopt.x
        tors = np.arctan2(np.sin(tors), np.cos(tors))
        tors = np.sqrt(tors @ tors)
        energy = xopt.fun + tors
        return energy

    def trim_rotamers(self, keep_idx: ArrayLike = None) -> None:
        """
        Remove rotamers with small weights from ensemble

        Parameters
        ----------
        keep_idx : ArrayLike[int]
            Indices of rotamers to keep. If None rotamers will be trimmed based off of `self.trim_tol`.
        """
        if keep_idx is None:
            arg_sort_weights = np.argsort(self.weights)[::-1]
            sorted_weights = self.weights[arg_sort_weights]
            cumulative_weights = np.cumsum(sorted_weights)
            cutoff = np.maximum(1, len(cumulative_weights[cumulative_weights < 1 - self.trim_tol]))
            keep_idx = arg_sort_weights[:cutoff]

        if len(self.weights) == len(self.internal_coords):
            self.internal_coords = self.internal_coords.copy()
            self.internal_coords.use_frames(keep_idx)

        self._coords = self._coords[keep_idx]
        self._dihedrals = self._dihedrals[keep_idx]
        self.weights = self.weights[keep_idx]
        if self.atom_energies is not None:
            self.atom_energies = self.atom_energies[keep_idx]

        # normalize weights
        self.weights /= self.weights.sum()

    @property
    def centroid(self):
        """Get the centroid of the whole rotamer ensemble."""
        return self._coords.mean(axis=(0, 1))

    def evaluate(self):
        """Place rotamer ensemble on protein site and recalculate rotamer weights."""

        # Calculate external energies
        energies = self.energy_func(self)

        # Calculate total weights (combining internal and external)
        self.weights, self.partition = scoring.reweight_rotamers(energies, self.temp, self.weights)

        # Remove low-weight rotamers from ensemble
        if self._do_trim:
            self.trim_rotamers()

    def save_pdb(self, name: str = None) -> None:
        """
        Save a pdb file containing the rotamer ensemble.

        Parameters
        ----------
        name : str | None
            Name of the file. If `None` self.name will be used.
        """
        io.save(name, self)

    @property
    def backbone(self):
        """Backbone coordinates of the spin label"""
        return np.squeeze(self._coords[0][self.backbone_idx])

    @property
    def aln(self):
        return np.squeeze(self._coords[0][self.aln_idx])


    @property
    def side_chain(self):
        return np.squeeze(self._coords[0][self.side_chain_idx])


    def get_lib(self, rotlib):
        """Parse backbone information from protein and fetch the appropriate rotamer library.

        Parameters
        ----------
            rotlib : str | dict
                The name of the rotlib or a dictionary containing all the information required for a rotamer library.
        Returns
        -------
        lib : dict
            Dictionary containing all underlying attributes of a rotamer library.
        """

        PhiSel, PsiSel = None, None
        if self.protein is not None:
            # get site backbone information from protein structure
            PhiSel = (
                self.protein.select_atoms(f"resid {self.site} and segid {self.chain}")
                .residues[0]
                .phi_selection()
            )
            PsiSel = (
                self.protein.select_atoms(f"resid {self.site} and segid {self.chain}")
                .residues[0]
                .psi_selection()
            )

        # Default to helix backbone if none provided
        Phi = (
            None
            if PhiSel is None
            else np.rad2deg(get_dihedral(PhiSel.positions))
        )
        Psi = (
            None
            if PsiSel is None
            else np.rad2deg(get_dihedral(PsiSel.positions))
        )
        self.Phi, self.Psi = Phi, Psi

        # Get library
        cwd = Path().cwd()

        was_none = True if rotlib is None else False
        rotlib = self.res if rotlib is None else rotlib

        # If the rotlib isn't a dict try to figure out where it is
        if not isinstance(rotlib, dict):
            trotlib = io.get_possible_rotlibs(rotlib, suffix='rotlib', extension='.npz', was_none=was_none)
            if trotlib:
                rotlib = trotlib
            elif rotlib not in SUPPORTED_RESIDUES:
                raise NameError(f'There is no rotamer library called {rotlib} in this directory or in chilife')

            lib = io.read_library(rotlib, Phi, Psi)

        # If rotlib is a dict, assume that dict is in the format that read_library`would return
        else:
            lib = rotlib

        # Copy mutable library values (mostly np arrays) so that instances do not share data
        lib = {key: value.copy() if hasattr(value, 'copy') else value for key, value in lib.items()}

        # Perform a sanity check to ensure every necessary entry is present
        if isinstance(rotlib, Path) and not all(x in lib for x in io.rotlib_formats[lib['format_version']]):
            raise ValueError('The rotamer library does not contain all the required entries for the format version')

        # Modify library to be appropriately used with self.__dict__.update
        self._rotlib = {key: value.copy() if hasattr(value, 'copy') and key != 'internal_coords' else value for key, value in lib.items()}

        lib['_coords'] = lib.pop('coords').copy()
        lib['_dihedrals'] = lib.pop('dihedrals').copy()
        if 'skews' not in lib:
            lib['skews'] = None
        return lib



    def init_protein(self):
        if isinstance(self.protein, (mda.AtomGroup, mda.Universe)):
            if not hasattr(self.protein.universe._topology, "altLocs"):
                self.protein.universe.add_TopologyAttr('altLocs', np.full(len(self.protein.universe.atoms), ""))

        # Position library at selected residue
        self.resindex = self.protein.select_atoms(self.selstr).resindices[0]
        self.segindex = self.protein.select_atoms(self.selstr).segindices[0]
        if self.ignore_waters:
            self._protein = self.protein.select_atoms("not (byres name OH2 or resname HOH)")
        else:
            self._protein = self.protein.atoms

        self.protein_tree = cKDTree(self._protein.atoms.positions)

        # Delete cached lennard jones parameters if they exist.
        if hasattr(self, 'ermin_ij'):
            del self.ermin_ij
            del self.eeps_ij


    @property
    def bonds(self):
        """Set of atom indices corresponding to the atoms that form covalent bonds"""

        if not hasattr(self, "_bonds"):
            icmask_map = {x: i for i, x in enumerate(self.ic_mask)}
            self._bonds = np.array([(icmask_map[a], icmask_map[b]) for a, b in self.internal_coords.bonds
                                   if a in icmask_map and b in icmask_map])
            
        return self._bonds

    @bonds.setter
    def bonds(self, inp):
        self._bonds = set(tuple(i) for i in inp)
        idxs = np.arange(len(self.atom_names))
        all_pairs = set(combinations(idxs, 2))
        self._non_bonded = all_pairs - self._bonds

    @property
    def non_bonded(self):
        """Set of atom indices corresponding to the atoms that are not covalently bonded. Also excludes atoms that
        have 1-n non-bonded interactions where `n=self._exclude_nb_interactions` . By default, 1-3 interactions are
        excluded"""
        if not hasattr(self, "_non_bonded"):
            pairs = {v.index: [path for path in self._graph.get_all_shortest_paths(v) if
                           len(path) <= (self._exclude_nb_interactions)] for v in self._graph.vs}
            pairs = {(a, c) for a in pairs for b in pairs[a] for c in b if a < c}
            all_pairs = set(combinations(range(len(self.atom_names)), 2))
            self._non_bonded = all_pairs - pairs

        return sorted(list(self._non_bonded))

    @non_bonded.setter
    def non_bonded(self, inp):
        self._non_bonded = set(tuple(i) for i in inp)
        idxs = np.arange(len(self.atom_names))
        all_pairs = set(combinations(idxs, 2))
        self._bonds = all_pairs - self._non_bonded

    def __len__(self):
        """Number of rotamers in the ensemble"""
        return len(self.coords)

    @property
    def clash_ori(self):
        """Location to use as 'center' when searching for atoms with potential clashes"""
        if isinstance(self._clash_ori_inp, (np.ndarray, list)):
            if len(self._clash_ori_inp) == 3:
                return self._clash_ori_inp
        elif isinstance(self._clash_ori_inp, str):
            if self._clash_ori_inp in ["cen", "centroid"]:
                return self.centroid
            elif (ori_name := self._clash_ori_inp.upper()) in self.atom_names:
                return np.squeeze(self._coords[0][ori_name == self.atom_names])
        else:
            raise ValueError(
                f"Unrecognized clash_ori option {self._clash_ori_inp}. Please specify a 3D vector, an "
                f"atom name or `centroid`"
            )

        return self._clash_ori

    @clash_ori.setter
    def clash_ori(self, inp):
        self._clash_ori_inp = inp

    @property
    def coords(self):
        """Cartesian coordinates of all rotamers in the ensemble."""
        return self._coords

    @coords.setter
    def coords(self, coords):
        # TODO: Add warning about removing isomers

        # Allow users to input a single rotamer
        coords = coords if coords.ndim == 3 else coords[None, :, :]

        # Check if atoms match
        if coords.shape[1] == len(self.side_chain_idx):
            tmp = np.array([self._coords[0].copy() for _ in coords])
            tmp[:, self.side_chain_idx] = coords
            coords = tmp

        if coords.shape[1] != len(self.atoms):
            raise ValueError('The number of atoms in the input array does not match the number of atoms of the residue')

        self._coords = coords
        self.internal_coords = self.internal_coords.copy()
        self.internal_coords.set_cartesian_coords(coords, self.ic_mask)

        # Check if they are all at the same site
        self._dihedrals = np.rad2deg([ic.get_dihedral(1, self.dihedral_atoms) for ic in self.internal_coords])

        # Apply uniform weights
        self.weights = np.ones(len(self._dihedrals))
        self.weights /= self.weights.sum()

    @property
    def dihedrals(self):
        """Values of the (mobile) dihedral angles of all rotamers in the ensemble"""
        return self._dihedrals

    @dihedrals.setter
    def dihedrals(self, dihedrals):

        dihedrals = dihedrals if dihedrals.ndim == 2 else dihedrals[None, :]
        if dihedrals.shape[1] != self.dihedrals.shape[1]:
            raise ValueError('The input array does not have the correct number of dihedrals')

        if len(dihedrals) != self.dihedrals.shape[0]:

            warnings.warn('WARNING: Setting dihedrals in this fashion will set all bond lengths and angles to that '
                          'of the first rotamer in the library effectively removing stereo-isomers from the ensemble. It '
                          'will also set all weights to .')

            # Apply uniform weights
            self.weights = np.ones(len(dihedrals))
            self.weights /= self.weights.sum()
            idxs = [0 for _ in range(len(dihedrals))]
        else:
            idxs = np.arange(len(dihedrals))

        self._dihedrals = dihedrals
        z_matrix = self.internal_coords.batch_set_dihedrals(idxs, np.deg2rad(dihedrals), 1, self.dihedral_atoms)
        self.internal_coords = self.internal_coords.copy()
        self.internal_coords.load_new(z_matrix)
        self._coords = self.internal_coords.protein.trajectory.coordinate_array.copy()[:, self.ic_mask]
        if self._match_backbone:
            self.backbone_to_site()



    @property
    def mx(self):
        """The rotation matrix to rotate a residue from the local coordinate frame to the current residue. The local
        coordinate frame is defined by the alignment method"""
        mx, ori = global_mx(*parse_backbone(self, kind="local"), method=self.alignment_method)
        return mx

    @property
    def origin(self):
        """Origin of the local coordinate frame"""
        return np.squeeze(self.aln[1])

    @property
    def CB(self):
        """The coordinates of the β-carbon atom of the RotamerEnsemble"""
        if 'CB' not in self.atom_names:
            raise ValueError("There is no CB atom in this side chain")
        cbidx = np.argwhere(self.atom_names == 'CB').flat[0]
        return self.coords[:, cbidx].mean(axis=0)

    @property
    def protein(self):
        return self._protein

    @protein.setter
    def protein(self, protein):

        if hasattr(protein, "atoms") and isinstance(protein.atoms, (mda.AtomGroup, MolecularSystemBase)):
            self._protein = protein
            self.init_protein()
            resname = self.protein.select_atoms(self.selstr)[0].resname
            self.nataa = nataa_codes.get(resname, resname)

        elif protein is None:
            self._protein = protein
        else:
            raise ValueError("The input protein must be an instance of MDAnalysis.Universe, MDAnalysis.AtomGroup, or "
                             "chilife.MolSys")

    def intra_fit(self):
        """Align all rotamers of the ensemble to the ensemble backbone"""

        target = self.aln

        tmx, tori = self.alignment_method(*target)
        bbs = np.squeeze(self.coords[:,self.aln_idx])
        mxs, oris = [np.array(x) for x in zip(*[self.alignment_method(*bb) for bb in bbs])]
        mxs = mxs.transpose(0, 2, 1) @ tmx

        self._coords = (self.coords - oris[:, None, :]) @ mxs + tori[None, None, :]
        self.ICs_to_site(tori, tmx)


    def get_sasa(self):
        """Calculate the solvent accessible surface area (SASA) of each rotamer in the protein environment."""

        atom_radii = self.energy_func.get_lj_rmin(self.atom_types)
        if self.protein is not None:
            environment_coords = self.protein.atoms[self.protein_clash_idx].positions
            environment_radii = self.energy_func.get_lj_rmin(self.protein.atoms[self.protein_clash_idx].types)
        else:
            environment_coords = np.empty((0, 3))
            environment_radii = np.empty(0)

        SASAs = nu_getsasa(self.coords, atom_radii, environment_coords, environment_radii)

        return np.array(SASAs)

    def set_dihedral_sampling_sigmas(self, value):
        """
        Helper function to assign new sigmas for off-rotamer sampling.

        Parameters
        ----------
        value : float | ArrayLike[float]
            The sigma value(s) to assign. A single float will assign all dihedrals of all rotamers to the same sigma
            (isotropic). An array with as many elements as dihedrals will assign different sigmas to each dihedral
            (Anisotropic). An array of dimensions n_rots by n_dihedrals will assign different anisotropic sigmas to
            different rotamers.
        """
        value = np.asarray(value)

        if value.shape == ():
            self.sigmas = (np.ones((len(self._weights), len(self.dihedral_atoms))) * value)
        elif len(value) == len(self.dihedral_atoms) and len(value.shape) == 1:
            self.sigmas = np.tile(value, (len(self._weights), 1))
        elif value.shape == (len(self._weights), len(self.dihedral_atoms)):
            self.sigmas = value.copy()
        else:
            raise ValueError('`dihedral_sigmas` must be a scalar, an array the length of the `self.dihedral atoms` or '
                             'an array with the shape of (len(self.weights), len(self.dihedral_atoms))')


def assign_defaults(kwargs):
    """
    Helper function to assign default values to kwargs that have not been explicitly assigned by the user. Also
    checks to make sure that the user provided kwargs are real kwargs.
    Parameters
    ----------
    kwargs : dict
        Dictionary of user supplied keyword arguments.

    Returns
    -------
    kwargs : dict
        Dictionary of user supplied keyword arguments augmented with defaults for all kwargs the user did not supply.
    """

    # Make all string arguments lowercase
    for key, value in kwargs.items():
        if isinstance(value, str):
            kwargs[key] = value.lower()

    # Default parameters
    defaults = {
        "protein_tree": None,
        "temp": 298,
        "clash_radius": None,
        "_clash_ori_inp": kwargs.pop("clash_ori", "cen"),
        "alignment_method": "bisect",
        "dihedral_sigmas": 35,
        "weighted_sampling": False,
        "eval_clash": True if not kwargs.get('minimize', False) else False,
        "use_H": False,
        '_match_backbone': True,
        "_exclude_nb_interactions": kwargs.pop('exclude_nb_interactions', 3),
        "_sample_size": kwargs.pop("sample", False),
        "energy_func": default_energy_func,
        "_minimize": kwargs.pop('minimize', False),
        "min_method": 'L-BFGS-B',
        "_do_trim": kwargs.pop('trim', True),
        "trim_tol": 0.005,
        "ignore_waters": True
    }

    # Overwrite defaults
    kwargs_filled = {**defaults, **kwargs}

    # Make sure there are no unused parameters
    if len(kwargs_filled) != len(defaults):
        raise TypeError(
            f"Got unexpected keyword argument(s): "
            f'{", ".join(key for key in kwargs if key not in defaults)}'
        )

    return kwargs_filled.items()


def guess_chain(protein, site):
    """
    Reads chain from protein or makes an educated guess on which chain a particular site resides on for a given
    Protein/Universe/AtomGroup.

    Parameters
    ----------
    protein : mda.Universe | mda.AtomGroup | chilife.MolSys
        The protein being labeled.
    site :  int
        The residue being labeled.
    Returns
    -------
    chain : str
        Best guess for the chain on which the selected residue resides.

    """
    site = int(site)
    if protein is None:
        chain = "A"
    elif len(set(protein.segments.segids)) == 1:
        chain = protein.segments.segids[0]
    elif np.isin(protein.residues.resids, site).sum() == 0:
        raise ValueError(f"Residue {site} is not present on the provided protein")
    elif np.isin(protein.residues.resnums, site).sum() == 1:
        chain = protein.select_atoms(f"resid {site}").segids[0]
    else:
        raise ValueError(
            f"Residue {site} is present on more than one chain. Please specify the desired chain"
        )
    return chain



import numpy as np
import pytest
import MDAnalysis as mda
import chilife

import matplotlib.pyplot as plt

r = np.linspace(1e-3, 3, 256)
eps = np.ones(len(r))
rmin = np.ones(len(r))
protein = mda.Universe("test_data/1ubq.pdb", in_memory=True)
lj_funcs = [chilife.get_lj_energy, chilife.get_lj_scwrl, chilife.get_lj_rep]
lj_ans = [np.array([-1.29407456, -1.34917747, -0.92307553,  0.09701479, -0.81320874,
                    -0.82519134, -0.81681929, -0.47411124, -0.64444505, -0.44549313]),
          np.array([-0.05528211, -0.14662196,  0.        , -0.20341   ,  0.        ,
                     0.        ,  0.        ,  0.        ,  0.        ,  0.        ]),
          np.array([0.66692039, 0.05328766, 0.01843082, 0.01921865, 0.00483313,
                    0.68429473, 0.00290177, 0.00411807, 1.25053399, 0.03108683])]


@pytest.mark.parametrize(('func', 'ans'), zip(lj_funcs, lj_ans))
def test_lj(func, ans):
    f = chilife.ljEnergyFunc(func)
    RL = chilife.RotamerEnsemble('TRP', 28, protein, energy_func=f, eval_clash=True)
    np.testing.assert_almost_equal(RL.atom_energies[5], ans, decimal=6)


@pytest.mark.parametrize('func',  lj_funcs)
def test_efunc(func):
    RL = chilife.RotamerEnsemble('TRP', 28, protein, eval_clash=False)
    f = chilife.ljEnergyFunc(func)
    test = f(RL)
    ans = np.load(f'test_data/{func.__name__}.npy')
    np.testing.assert_almost_equal(test, ans, decimal=5)


@pytest.mark.parametrize('func',  lj_funcs)
def test_efunc_dlabel(func):
    dSL = chilife.dSpinLabel('DHC', (28, 32), protein, eval_clash=False, rotlib='test_data/DHC')
    f = chilife.ljEnergyFunc(func)
    test = f(dSL)
    ans = np.load(f'test_data/d{func.__name__}.npy')
    np.testing.assert_almost_equal(test, ans, decimal=4)


def test_molar_gas_constant():
    np.testing.assert_almost_equal(chilife.scoring.GAS_CONST, 1.987204258640832e-3, decimal=10)


def test_get_lj_case_sensitivity():
    ff = chilife.ljEnergyFunc()
    x = ff.get_lj_rmin(['CA', 'Ca', 'ca'])
    assert np.all(x == 1.367)


import hashlib, os, pickle
from itertools import chain
import pytest
import numpy as np
import MDAnalysis as mda
import chilife as xl

pdbids = ["1ubq", "1a2w", '1az5']
ubq = mda.Universe("test_data/1ubq.pdb", in_memory=True)
mbp = mda.Universe("test_data/2klf.pdb", in_memory=True)

ubqIC = xl.MolSysIC.from_atoms(ubq)
mbpIC = xl.MolSysIC.from_atoms(mbp)


gd_kwargs = [
    {"resi": 28, "atom_list": ["C", "N", "CA", "C"]},
    {"resi": 28, "atom_list": [["C", "N", "CA", "C"], ["N", "CA", "C", "N"]]},
]
gd_ans = [-1.1540451, np.array([-1.1540451, -0.6653183])]


def test_from_prot():

    with np.load('test_data/ubq_zmats.npz') as f:
        zmat_ans = f['zmat']
        zmat_idx_ans = f['zmat_idxs']


    non_nan_idx = ubqIC.z_matrix_idxs > 0
    np.testing.assert_equal(ubqIC.z_matrix_idxs[non_nan_idx], zmat_idx_ans[non_nan_idx])
    np.testing.assert_almost_equal(ubqIC.trajectory.coordinate_array[:, ubqIC.non_nan_idxs],
                                   zmat_ans[:, ubqIC.non_nan_idxs], decimal=6)

    for i, idxs in enumerate(ubqIC.z_matrix_idxs):
        idxt = [val for val in idxs if val != -2147483648]
        assert i == idxt[0]


def test_from_prot_traj():

    with np.load('test_data/mbp_zmats.npz') as f:
        zmat_ans = f['zmat']
        zmat_idx_ans = f['zmat_idxs']

    np.testing.assert_equal(mbpIC.z_matrix_idxs[mbpIC.non_nan_idxs], zmat_idx_ans[mbpIC.non_nan_idxs])
    np.testing.assert_almost_equal(mbpIC.trajectory.coordinate_array[:, mbpIC.non_nan_idxs],
                                   zmat_ans[:, mbpIC.non_nan_idxs], decimal=6)


@pytest.mark.parametrize("pdbid", pdbids)
def test_get_internal_coordinates(pdbid):
    protein = mda.Universe(f"test_data/{pdbid}.pdb", in_memory=True).select_atoms("protein and not altloc B")
    ICs = xl.MolSysIC.from_atoms(protein, ignore_water=False)
    np.testing.assert_almost_equal(ICs.coords, protein.atoms.positions, decimal=4)

    protein = xl.MolSys.from_pdb(f"test_data/{pdbid}.pdb").select_atoms("protein and not altloc B")
    ICs = xl.MolSysIC.from_atoms(protein, ignore_water=False)
    np.testing.assert_almost_equal(ICs.coords, protein.atoms.positions, decimal=4)


def test_copy():
    my_copy = ubqIC.copy()

    assert np.all(my_copy.z_matrix_idxs == ubqIC.z_matrix_idxs)
    assert my_copy.z_matrix_idxs is not ubqIC.z_matrix_idxs

    assert np.all(ubqIC.trajectory.coordinate_array == my_copy.trajectory.coordinate_array)
    assert ubqIC.trajectory.coordinate_array is not my_copy.trajectory.coordinate_array
    for v1, v2 in zip(my_copy.chain_operators.values(), ubqIC.chain_operators.values()):
        for v11, v22 in zip(v1.values(), v2.values()):
            assert np.all(v11 == v22)


def test_save_pdb():
    protein = mda.Universe("test_data/alphabetical_peptide.pdb").select_atoms("protein")

    uni_ics = xl.MolSysIC.from_atoms(protein)
    xl.save("alphabetical_peptide.pdb", uni_ics)

    with open("alphabetical_peptide.pdb", "r") as f:
        test = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    with open("test_data/alphabetical_peptide.pdb", "r") as f:
        truth = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    os.remove("alphabetical_peptide.pdb")
    assert test == truth


def test_has_clashes():
    local_ubqIC = ubqIC.copy()
    assert not local_ubqIC.has_clashes()
    local_ubqIC.set_dihedral(np.pi / 2, 35, ["N", "CA", "C", "N"])
    assert local_ubqIC.has_clashes()


def test_set_dihedral():
    local_ubqIC = ubqIC.copy()
    local_ubqIC.set_dihedral([-np.pi / 2, np.pi / 2], 72, [["C", "N", "CA", "C"], ["N", "CA", "C", "N"]])

    ans = mda.Universe("test_data/test_icset.pdb")
    np.testing.assert_almost_equal(ans.atoms.positions, local_ubqIC.coords, decimal=3)


def test_set_dihedral2():
    lys = mda.Universe('../src/chilife/data/rotamer_libraries/residue_pdbs/lys.pdb')

    ICs = xl.MolSysIC.from_atoms(lys)
    ICs.set_dihedral(np.pi / 2, 1, ["N", "CA", "CB", "CG"])

    print('DSF')


@pytest.mark.parametrize(["inp", "ans"], zip(gd_kwargs, gd_ans))
def test_get_dihedral(inp, ans):
    ubqIC = xl.MolSysIC.from_atoms(ubq)
    dihedral = ubqIC.get_dihedral(**inp)
    np.testing.assert_almost_equal(dihedral, ans, decimal=4)



def test_polypro():
    polypro = mda.Universe("test_data/PPII_Capped.pdb")
    polyproIC = xl.MolSysIC.from_atoms(polypro)
    np.testing.assert_equal(polyproIC.z_matrix_names[12], ['CD', 'CG', 'CB', 'CA'])


def test_PRO():
    pro = mda.Universe("../src/chilife/data/rotamer_libraries/residue_pdbs/pro.pdb")
    pro_ic = xl.MolSysIC.from_atoms(pro)

    np.testing.assert_equal(pro_ic.z_matrix_names[6], ["CD", "CG", "CB", "CA"])

def test_PRO2():
    zmat_idxs = ubqIC.chain_res_name_map['A', 37, 'N', 'CA']
    np.testing.assert_equal(ubqIC.z_matrix_names[zmat_idxs[0]], ["C", "CA", "N", "C"] )


def test_set_coords():
    R1A = mda.Universe("test_data/R1A.pdb")
    R1A_IC = xl.MolSysIC.from_atoms(R1A)
    R1A_IC_c = R1A_IC.copy()
    R1A_IC_c.set_dihedral([np.pi/2, -np.pi/2, np.pi/2], 1, [['N', 'CA', 'CB', 'SG' ],
                                                            ['CA', 'CB', 'SG', 'SD'],
                                                            ['CB', 'SG', 'SD', 'CE']])

    coords = R1A_IC_c.coords

    R1A_IC.coords = coords
    np.testing.assert_allclose(R1A_IC.z_matrix, R1A_IC_c.z_matrix, rtol=1e-5)
    np.testing.assert_almost_equal(R1A_IC.coords, R1A_IC_c.coords, decimal=6)


def test_nonbonded():
    with open('test_data/ic_nb.pkl', 'rb') as f:
        nb_ans = pickle.load(f)

    np.testing.assert_equal(ubqIC.nonbonded, nb_ans)


def test_get_zmat_idxs():
    R1A = mda.Universe("test_data/R1A.pdb")
    R1A_IC = xl.MolSysIC.from_atoms(R1A)
    idxs = R1A_IC.get_z_matrix_idxs(1, ['CB', 'SG', 'SD', 'CE'])

    np.testing.assert_equal(idxs, 7)

    with pytest.raises(RuntimeError):
        idxs = R1A_IC.get_z_matrix_idxs(1, ['CE', 'SD', 'SG', 'CB'])


def test_phi_idxs():
    idxs = ubqIC.phi_idxs(range(4, 11,))
    vals = ubqIC.z_matrix[idxs, -1]
    ans = [ubqIC.get_dihedral(i, ('C', 'N', 'CA', 'C')) for i in range(4, 11)]

    np.testing.assert_almost_equal(vals, ans, decimal=5)

    idx = ubqIC.phi_idxs(1)
    assert len(idx) == 0


def test_psi_idxs():
    idxs = ubqIC.psi_idxs(range(4, 11,))
    vals = ubqIC.z_matrix[idxs, -1]
    ans = [ubqIC.get_dihedral(i, ('N', 'CA', 'C', 'N')) for i in range(4, 11)]

    np.testing.assert_almost_equal(vals, ans, decimal=4)

    idx = ubqIC.psi_idxs(76)
    assert len(idx) == 0

    idxs = ubqIC.psi_idxs()
    assert 0 not in idxs




def test_phi_psi_idxs_multichain():
    prot = xl.fetch('1a2w').select_atoms('protein')
    ICs = xl.MolSysIC.from_atoms(prot)

    with pytest.raises(ValueError):
        ICs.psi_idxs(10)

    with pytest.raises(ValueError):
        ICs.psi_idxs(10)

    A = ICs.psi_idxs(10, chain='A')
    B = ICs.phi_idxs(10, chain='A')

    assert A[0] == 80
    assert B[0] == 71

    A = ICs.psi_idxs(10, chain='B')
    B = ICs.phi_idxs(10, chain='B')

    assert A[0] == 1031
    assert B[0] == 1022

def test_chi_idxs():
    idxs = ubqIC.chi_idxs(range(4, 11, ))
    idxs = list(chain.from_iterable(idxs))
    vals = ubqIC.z_matrix[idxs, -1]
    ans = np.array([-1.0589304 ,  1.70940745, -3.13784337, -3.06411982,  3.0752275 ,
                    -3.00272059, -3.04088092,  1.33890903, -1.18974721, -2.89442134,
                     1.23405945])

    np.testing.assert_almost_equal(vals, ans, decimal=6)

def test_ic_pref_dihe():

    mol = mda.Universe('test_data/test_ic_pref_dihe.pdb', in_memory=True)
    dih = [['N', 'CA', 'CB', 'CB2'],
           ['CA', 'CB', 'CB2', 'NG'],
           ['ND', 'CE3', 'CZ3', 'C31']]

    IC = xl.MolSysIC.from_atoms(mol, preferred_dihedrals=dih)
    IC.set_dihedral(np.pi / 2, 1, ['ND', 'CE3', 'CZ3', 'C31'])

    sister_dihe_atom_coords = IC.coords[IC.atom_names == 'C36'].flat
    np.testing.assert_almost_equal(sister_dihe_atom_coords, [2.1495337, -3.4064763, -0.9388056], decimal=6)


def test_pickle():
    mol1 = mda.Universe('test_data/PPII_Capped.pdb', in_memory=True)
    mol2 = xl.MolSys.from_pdb('test_data/PPII_Capped.pdb')
    for mol in (mol1, mol2):
        prot2 = mol

        with open('tmp.pkl', 'wb') as f:
            pickle.dump(prot2, f)

        del prot2
        del mol

        with open('tmp.pkl', 'rb') as f:
            prot2 = pickle.load(f)

    os.remove('tmp.pkl')


def test_chain_operators():
    LYS = mda.Universe('../src/chilife/data/rotamer_libraries/residue_pdbs/lys.pdb')
    pic = xl.MolSysIC.from_atoms(LYS)

    mx = np.array([[0.38281548,  0.9238248,   0.],
                   [0.92382485, -0.38281548,  0.],
                   [0.,          0., -0.99999994]])
    ori = np.array([-0.449, -3.572, -1.666])

    np.testing.assert_almost_equal(pic.chain_operators[0]['mx'], mx)
    np.testing.assert_almost_equal(pic.chain_operators[0]['ori'], ori)


def test_iter():
    zdata = []
    cdata = []
    for ic in mbpIC:
        cdata.append(ic.coords.copy())
        zdata.append(ic.z_matrix.copy())

    zans = []
    cans = []
    for ts in mbpIC.trajectory:
        cans.append(mbpIC.coords.copy())
        zans.append(mbpIC.z_matrix.copy())

    assert np.sum(np.abs(cdata[0] - cdata[1])) > 0
    np.testing.assert_equal(cdata, cans)
    np.testing.assert_equal(zdata, zans)

def test_load_new():
    pass

def test_use_frames():
    pass


from functools import partial
import numpy as np
from scipy.spatial.distance import cdist

from .scoring import ljEnergyFunc, get_lj_energy
from .RotamerEnsemble import RotamerEnsemble


class SpinLabel(RotamerEnsemble):
    """
    A RotamerEnsemble made from a side chain with one or more unpaired electrons.

    Parameters
    ----------
    label: str
        Name of desired spin label, e.g. R1A.
    site: int
        MolSys residue number to attach spin label to.
    chain: str
        MolSys chain identifier to attach spin label to.
    protein: MDAnalysis.Universe, MDAnalysis.AtomGroup
        Object containing all protein information (coords, atom types, etc.)
    protein_tree: cKDtree
        k-dimensional tree object associated with the protein coordinates.
    """

    def __init__(self, label, site=None, protein=None, chain=None, rotlib=None, **kwargs):

        # Override RotamerEnsemble default of not evaluating clashes
        if not kwargs.get('minimize', False):
            kwargs.setdefault("eval_clash", True)

        super().__init__(label, site, protein=protein, chain=chain, rotlib=rotlib, **kwargs)

        self.label = label

        # Parse spin delocalization information
        sa_mask = np.isin(self.spin_atoms, self.atom_names)
        self.spin_atoms = self.spin_atoms[sa_mask]
        self.spin_weights = self.spin_weights[sa_mask]
        self.spin_idx = np.argwhere(np.isin(self.atom_names, self.spin_atoms))
        self.spin_idx.shape = -1

    @property
    def spin_coords(self):
        """get the spin coordinates of the rotamer ensemble"""
        return np.squeeze(self._coords[:, self.spin_idx, :])

    @property
    def spin_centers(self):
        """get the spin center of the rotamers in the ensemble"""
        if np.any(self.spin_idx):
            spin_centers = np.average(self._coords[:, self.spin_idx, :], weights=self.spin_weights, axis=1)
        else:
            spin_centers = np.array([])
        return np.atleast_2d(np.squeeze(spin_centers))

    @property
    def spin_centroid(self):
        """Average location of all the label's `spin_coords` weighted based off of the rotamer weights"""
        return np.average(self.spin_centers, weights=self.weights, axis=0)


    @classmethod
    def from_mmm(cls, label, site=None, protein=None, chain=None, **kwargs):
        """Create a SpinLabel object using the default MMM protocol with any modifications passed via kwargs"""

        MMM_maxdist = {
            "R1M": 9.550856367392733 + 4,
            "R7M": 9.757254987175209 + 4,
            "V1M": 8.237071322458029 + 4,
            "M1M": 8.985723827323680 + 4,
            "I1M": 12.952083029729994 + 4,
        }


        clash_radius = kwargs.pop("clash_radius", MMM_maxdist.get(label, None))
        alignment_method = kwargs.pop("alignment_method", "mmm")
        clash_ori = kwargs.pop("clash_ori", "CA")
        energy_func = kwargs.pop('energy_func', ljEnergyFunc(get_lj_energy, 'uff', forgive=0.5, cap=np.inf))
        use_H = kwargs.pop("use_H", True)

        # Calculate the SpinLabel
        SL = SpinLabel(
            label,
            site,
            protein,
            chain,
            alignment_method=alignment_method,
            clash_radius=clash_radius,
            clash_ori=clash_ori,
            energy_func=energy_func,
            use_H=use_H,
            **kwargs,
        )

        return SL

    @classmethod
    def from_wizard(
        cls,
        label,
        site=None,
        protein=None,
        chain=None,
        to_find=200,
        to_try=10000,
        vdw=2.5,
        clashes=5,
        **kwargs,
    ):
        """Create a SpinLabel object using the default MTSSLWizard protocol with any modifications passed via kwargs"""

        prelib = cls(label, site, protein, chain, eval_clash=False, **kwargs)
        prelib.sigmas = np.array([])

        coords = np.zeros((to_find, len(prelib.atom_names), 3))
        internal_coords = []
        i = 0
        if protein is not None:
            protein_clash_idx = prelib.protein_tree.query_ball_point(prelib.centroid, 19.0)
            protein_clash_idx = [idx for idx in protein_clash_idx if idx not in prelib.clash_ignore_idx]

        a, b = [list(x) for x in zip(*prelib.non_bonded)]
        for _ in range(np.rint(to_try / to_find).astype(int)):
            sample, _, internal_sample = prelib.sample(n=to_find, off_rotamer=True, return_dihedrals=True)

            # Evaluate internal clashes
            dist = np.linalg.norm(sample[:, a] - sample[:, b], axis=2)
            sidx = np.atleast_1d(np.squeeze(np.argwhere(np.all(dist > 2, axis=1))))
            if len(sidx) == 0:
                continue

            sample = sample[sidx, ...]
            internal_sample.use_frames(sidx)
            if protein is not None:
                # Evaluate external clashes
                dist = cdist(
                    sample[:, prelib.side_chain_idx].reshape(-1, 3),
                    prelib.protein_tree.data[protein_clash_idx],
                )
                dist = dist.reshape(len(sidx), -1)
                nclashes = np.sum(dist < vdw, axis=1)
                sidx = np.atleast_1d(np.squeeze(np.argwhere(nclashes < clashes)))
            else:
                sidx = np.arange(len(sample))

            if len(sidx) == 0:
                continue

            if i + len(sidx) >= to_find:
                sidx = sidx[: to_find - i]

            coords[i: i + len(sidx)] = sample[sidx]
            internal_sample.use_frames(sidx)

            internal_coords.append(internal_sample.trajectory.coordinate_array.copy())
            i += len(sidx)

            if i == to_find:
                break

        coords = coords[coords.sum(axis=(1, 2)) != 0]
        z_matrix = np.concatenate(internal_coords) if len(internal_coords) > 0 else None
        if z_matrix is not None:
            internal_sample.trajectory.load_new(z_matrix)
            internal_sample.set_cartesian_coords(coords, prelib.ic_mask)

            prelib.internal_coords = internal_sample
            prelib._dihedrals = np.rad2deg([ic.get_dihedral(1, prelib.dihedral_atoms) for ic in prelib.internal_coords])
        else:
            prelib.internal_coords = None
            prelib._dihedrals = np.array([[]])

        prelib._coords = coords
        prelib.weights = np.ones(len(coords))
        prelib.weights /= prelib.weights.sum()
        return prelib

    def __str__(self):
        return (super().__str__() +
                f"  spin atoms:\n    {self.spin_atoms}")



    def _base_copy(self, rotlib=None):
        return SpinLabel(self.res, self.site, rotlib=rotlib, chain=self.chain)



from typing import List, Tuple, Set
from itertools import combinations, product
from operator import itemgetter

import numpy as np
from numpy.typing import ArrayLike
from scipy.spatial import cKDTree

import igraph as ig

from .globals import bond_hmax_dict
from .math_utils import simple_cycle_vertices

class Topology:
    """
    Topology class

    Parameters
    ----------
    mol : MDAnalysis.Universe, MDAnalysis.AtomGroup, chiLife.MolecularSystemBase
        Molecular system from which to define a topology.
    bonds : ArrayLike
        Array of tuples defining all the bonds of the molecule.
    kwargs : dict
        Additional keyword arguments, usually used to speed up construction by providing precomputed values for the
        topology attributes.

        graph : igraph.Graph
            A Graph object of the  molecule topology.
        angles : ArrayLike
            Array defining all bond-angles of the molecule.
        dihedrals : ArrayLike
            Array defining all dihedral angles of the molecule.

    """


    def __init__(self, mol, bonds, **kwargs):
        mol = mol.atoms
        self.atoms = mol.atoms
        self.atom_names = self.atoms.names
        self.atom_idxs = np.arange(len(mol))
        self.bonds = bonds
        self.bonds_any_atom = {}
        for b in self.bonds:
            for at in b:
                self.bonds_any_atom.setdefault(at, []).append(b)

        self.graph = kwargs.get('graph', self._make_graph())
        self.angles = kwargs.get('angles', get_angle_defs(self.graph))
        self.angles_any_atom = {}
        for a in self.angles:
            for at in a:
                self.angles_any_atom.setdefault(at, []).append(a)

        self.dihedrals = kwargs.get('dihedrals', get_dihedral_defs(self.graph))
        self.degree = []
        self.dihedrals_by_bonds = {}
        self.dihedrals_by_atoms = {}
        self.dihedrals_any_atom = {}
        self.dihedrals_by_resnum = {}

        for dihe in self.dihedrals:
            b, c, e = dihe[1:]
            bc_list = self.dihedrals_by_bonds.setdefault((b, c), [])
            bc_list.append(dihe)

            e_list = self.dihedrals_by_atoms.setdefault(e, [])
            e_list.append(dihe)

            for at in dihe:
                self.dihedrals_any_atom.setdefault(at, []).append(dihe)

            n1, n2, n3, n4 = self.atom_names[list(dihe)]
            r1 = self.atoms[c].resnum
            c1 = self.atoms[c].segid
            self.dihedrals_by_resnum[c1, r1, n1, n2, n3, n4] = dihe

    @property
    def ring_idxs(self):
        """Indices of atoms that are a part of one or more rings."""
        return simple_cycle_vertices(self.graph)

    @property
    def has_rings(self):
        if self.ring_idxs == []:
            return False
        else:
            return True

    def get_zmatrix_dihedrals(self):
        """
        Get the dihedral definitions for the z-matrix.

        Returns
        -------
        zmatrix_dihedrals : ArrayLike

        """
        zmatrix_dihedrals = []
        hold = []
        for key in self.atom_idxs:
            if key not in self.dihedrals_by_atoms:
                hold.append(key)
                continue
            elif 0 < len(hold) < 3:
                hold.append(key)
                continue
            else:
                for i, elem in enumerate(hold):
                    zmatrix_dihedrals.append(hold[:i + 1])
                hold = []

            runner_ups = []
            for dihe in self.dihedrals_by_atoms[key]:
                if dihe[0] < dihe[1] < dihe[2] < dihe[3]:
                    zmatrix_dihedrals.append(list(dihe))
                    break
                elif max(dihe[:3]) < dihe[3]:
                    runner_ups.append(dihe)
            else:
                if len(runner_ups) > 0:
                    minrup = np.argmin(np.sum(np.abs(np.diff(runner_ups, axis=1)), axis=1))
                    zmatrix_dihedrals.append(list(runner_ups[minrup]))

        return zmatrix_dihedrals

    def _make_graph(self):
        """Create an igraph.Graph form the topology"""
        return ig.Graph(n=len(self.atom_idxs), edges=self.bonds)

    def update_resnums(self):
        """Update the residue numbers of each atom with respect to the dihedral that they belong to."""
        self.dihedrals_by_resnum = {}
        for dihe in self.dihedrals:
            c = dihe[2]
            n1, n2, n3, n4 = self.atom_names[list(dihe)]
            r1 = self.atoms[c].resnum
            c1 = self.atoms[c].segid
            self.dihedrals_by_resnum[c1, r1, n1, n2, n3, n4] = dihe


def get_angle_defs(graph: ig.Graph) -> Tuple[Tuple[int, int, int]]:
    """
    Get all angle definitions for the topology defined by the graph.

    Parameters
    ----------
    graph: igraph.Graph
        A graph of the molecular topology.

    Returns
    -------
    angles : Tuple[Tuple[int, int, int]]
        Tuple containing tuples defining all angles of the molecule/molecular system.
    """
    angles = []
    for node in graph.vs.indices:
        neighbors = tuple(graph.neighbors(node))
        if len(neighbors) > 1:
            atom_angles = [(min(c), node, max(c)) for c in combinations(neighbors, 2)]
            angles += [tuple(a) for a in atom_angles]

    return tuple(angles)


def get_dihedral_defs(graph):
    """
    Get all dihedral definitions for the topology defined by the graph.

    Parameters
    ----------
    graph: igraph.Graph
        A graph of the molecular topology.

    Returns
    -------
    dihedrals : Tuple[Tuple[int, int, int, int]]
        Tuple containing tuples defining all dihedrals of the molecule/molecular system.
    """

    dihedrals = []
    for a, b in graph.get_edgelist():
        a_neighbors = tuple(graph.neighbors(a))
        b_neighbors = tuple(graph.neighbors(b))

        if len(a_neighbors) > 1 and len(b_neighbors) > 1:

            bond_dihedrals = [(aa, a, b, bb) if aa < bb else (bb, b, a, aa)
                                             for aa, bb in product(a_neighbors, b_neighbors)
                                             if all(idx not in (aa, bb) for idx in (a, b)) and aa != bb]

            dihedrals += [tuple(a) for a in bond_dihedrals]

    return tuple(dihedrals)


def get_min_topol(lines: List[List[str]],
                  forced_bonds: set = None) -> Set[Tuple[int, int]]:
    """ Git the minimum topology shared by all the states/models a PDB ensemble. This is to ensure a consistent
    internal coordinate system between all conformers of an ensemble even when there are minor differences in topology.
    e.g. when dHis-Cu-NTA has the capping ligand in different bond orientations.

    Parameters
    ----------
    lines : List[List[str]]
        List of lists corresponding to individual states/models of a pdb file. All models must have the same stoma in
        the same order and only the coordinates should differ.
    forced_bonds : set
        A set of bonds to that must be used regardless even if the bond lengths are not physically reasonable.
    Returns
    -------
    minimal_bond_set : Set[Tuple[int, int]]
        A set of tuples holding the indices of atom pairs which are thought to be bonded in all states/models.
    """
    bonds_list = []
    if isinstance(lines[0], str):
        lines = [lines]

    # Get bonds for all structures
    for struct in lines:
        coords = np.array([(line[30:38], line[38:46], line[46:54]) for line in struct], dtype=float)
        atypes = np.array([line[76:78].strip() for line in struct])
        pairs = guess_bonds(coords, atypes)
        bonds = set(tuple(pair) for pair in pairs)
        bonds_list.append(bonds)

    # Get the shared bonds between all structures.
    minimal_bond_set = set.intersection(*bonds_list)
    # Include any forced bonds
    if forced_bonds is not None:
        minimal_bond_set |= forced_bonds

    return minimal_bond_set


def guess_bonds(coords: ArrayLike, atom_types: ArrayLike) -> np.ndarray:
    """ Given a set of coordinates and their atom types (elements) guess the bonds based off an empirical metric.

    Parameters
    ----------
    coords : ArrayLike
        Array of three-dimensional coordinates of the atoms of a molecule or set of molecules for which you would like
        to guess the bonds of.
    atom_types : ArrayLike
        Array of the element symbols corresponding to the atoms of ``coords``

    Returns
    -------
    bonds : np.ndarray
        An array of the atom index pairs corresponding to the atom pairs that are thought ot form bonds.
    """
    atom_types = np.array([a.title() for a in atom_types])
    kdtree = cKDTree(coords)
    pairs = kdtree.query_pairs(4., output_type='ndarray')
    pair_names = [tuple(x) for x in atom_types[pairs].tolist()]
    bond_lengths = itemgetter(*pair_names)(bond_hmax_dict)
    a_atoms = pairs[:, 0]
    b_atoms = pairs[:, 1]

    dist = np.linalg.norm(coords[a_atoms] - coords[b_atoms], axis=1)
    bonds = pairs[dist < bond_lengths]
    sorted_args = np.lexsort((bonds[:, 0], bonds[:, 1]))
    return bonds[sorted_args]


def neighbors(edges, node):
    """
    Given a graph defined by edges and a node, find all neighbors of that node.

    Parameters
    ----------
    edges : ArrayLike
        Array of tuples defining all edges between nodes
    node : int
        The node of the graph for which to find neighbors.

    Returns
    -------
    nbs : ArrayLike
        Neighbor nodes.
    """
    nbs = []
    for edge in edges:
        if node not in edge:
            continue
        elif node == edge[0]:
            nbs.append(edge[1])
        elif node == edge[1]:
            nbs.append(edge[0])
    return nbs


def modified_bfs_edges(edges, root, bb_idxs):
    """
    Breadth first search of nodes given a set of edges
    Parameters
    ----------
    edges : ArrayLike
        Array of tuples defining edges between nodes.
    root : int
        Starting (root) node to begin the breadth first search at.

    Yields
    ------
    parent : int
        The node from which the children node stem
    child: List[int]
        All children node of parent.
    """
    nodes = np.unique(edges)

    depth_limit = len(nodes)
    seen = {root}

    n = len(nodes)
    depth = 0
    neigh = neighbors(edges, root)
    # Prioritize side chains
    neigh1 = [n for n in neigh if n not in bb_idxs]
    neigh2 = [n for n in neigh if n in bb_idxs]

    for neigh in neigh1, neigh2:
        next_parents_children = [(root, neigh)]
        while next_parents_children and depth < depth_limit:
            this_parents_children = next_parents_children
            next_parents_children = []
            for parent, children in this_parents_children:
                for child in children:
                    if child not in seen:
                        seen.add(child)
                        next_parents_children.append((child, neighbors(edges, child)))
                        yield parent, child
                if len(seen) == n:
                    return
            depth += 1
